<?php
/**
 * Project: 
 * DZIRCMS Content Management System
 * 
 * Description: 
 * Class for form field validation
 * 
 * Requirements: 
 * @con_php_req 7.4
 * 
 * 
 * @package     DZIRCMS Backend classes
 * @version     1.0.0
 * @author      René Mansveld
 * @copyright   Spider IT Deutschland
 * @link        https://www.spider-it.de
 * @link        https://www.dzir.org/cms/
 * @since       file available since drugCMS release 2.0.2
 * 
 * {@internal 
 *   created    2014-07-17
 * 
 *   modified:  2022-09-02 René Mansveld - Upgraded to PHP7.4
 *                                       - Added funktionallity
 * 
 * }}
 */

class FormValidation {
    
    /**
     * function isExistingEmailAddress()
     * 
     * Checks if the email address exists by asking it's mailserver
     * 
     * @param string $email E-Mail address to validate
     * 
     * @return bool Existance of the given address
     */
    public static function isExistingEmailAddress(string $email):bool {
        # Prepare a sender address for validation
        $sender = 'info@dzir.org';
        # Do the validation
        if (!is_array($email)) {
            $email = array($email);
        }
        $oValidator = new SMTP_validateEmail();
        $ret = $oValidator->validate($email, $sender);
        unset($oValidator);
        return $ret[$email[0]];
    }
    
    /**
     * function isExistingUrl()
     * 
     * Checks if the url exists by calling it
     * 
     * @param string $url Url to check
     * 
     * @return bool Existance of the given url
     */
    public static function isExistingUrl(string $url):bool {
        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);
        curl_setopt($ch, CURLOPT_HEADER, 1);         # get the header
        curl_setopt($ch, CURLOPT_NOBODY, 1);         # and *only* get the header
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1); # get the response as a string from curl_exec(), rather than echoing it
        curl_setopt($ch, CURLOPT_FRESH_CONNECT, 1);  # don't use a cached version of the url
        if (curl_exec($ch)) {
            curl_close($ch);
            return true;
        }
        return false;
    }

    /**
     * function isValidBic()
     * 
     * Checks if the value is numeric and currency
     * 
     * @param string $bic Value to check
     * @return bool Is the value a valid currency
     */
    public static function isValidBic(string $bic):bool {
        $ret = true;
        if ((strlen($bic) < 8) || (strlen($bic) > 11)) {
            $ret = false;
        }
        return $ret;
    }

    /**
     * function isValidCreditCard()
     * 
     * Checks if the value is numeric and currency
     * 
     * @param string $cardNumber Value to check
     * @return bool Is the value a valid currency
     */
    public static function isValidCreditCard(string $cardNumber):bool {
        $ret = true;
        $cards = array(
            'visa' => '(4\d{12}(?:\d{3})?)',
            'amex' => '(3[47]\d{13})',
            'jcb' => '(35[2-8][89]\d\d\d{10})',
            'maestro' => '((?:5020|5038|6304|6579|6761)\d{12}(?:\d\d)?)',
            'solo' => '((?:6334|6767)\d{12}(?:\d\d)?\d?)',
            'mastercard' => '(5[1-5]\d{14})',
            'switch' => '(?:(?:(?:4903|4905|4911|4936|6333|6759)\d{12})|(?:(?:564182|633110)\d{10})(\d\d)?\d?)',
            'discover' => '([6011\d{12}|[622126-622925]\d{10}|[644-649]\d{13}|65\d{14}])',
            'dinersclub' => "(3(?:(0[0-5]{1}[0-9]{11})|(6[0-9]{12})|(8[0-9]{12})))"
        );
        $matches = array();
        $pattern = '#^(?:' . implode('|', $cards) . ')$#';
        $result = preg_match($pattern, str_replace(' ', '', $cardNumber), $matches);
        $ret = ($result > 0);

        if ($ret) {
            /* Luhn algorithm number checker - (c) 2005-2008 shaman - www.planzero.org *
             * This code has been released into the public domain, however please      *
             * give credit to the original author where possible.                      */

            # Strip any non-digits (useful for credit card numbers with spaces and hyphens)
            $number = preg_replace('/\D/', '', $cardNumber);

            # Set the string length and parity
            $number_length = strlen($number);
            $parity = ($number_length % 2);

            # Loop through each digit and do the maths
            $total = 0;
            for ($i = 0; $i < $number_length; $i ++) {
                $digit = $number[$i];

                # Multiply alternate digits by two
                if (($i % 2) == $parity) {
                    $digit *= 2;

                    # If the sum is two digits, add them together (in effect)
                    if ($digit > 9) {
                        $digit -= 9;
                    }
                }
                # Total up the digits
                $total += $digit;
            }

            # If the total mod 10 equals 0, the number is valid
            $ret = (($total > 0) && (($total % 10) == 0));

        }

        return $ret;
    }

    /**
     * function isValidCurrency()
     * 
     * Checks if the value is numeric and currency
     * 
     * @param string $number Value to check
     * @return bool Is the value a valid currency
     */
    public static function isValidCurrency(string $number):bool {
        $ret = false;
        $match = array();
        if (preg_match("/\b^[-+]?\d{1,3}(?:,?\d{3})*(?:\.\d{2})?\b/", $number, $match)) {
            $ret = ($match[0] == $number);
        }
        return $ret;
    }
    
    /**
     * function isValidDate()
     * 
     * Checks if the given date matches the given format and the validity of the date.
     * 
     * @param string $date The date to evaluate
     * @param string $format Format of the date.
     *                       Any combination of <i>mm<i>, <i>dd<i>, <i>yyyy<i>
     *                       with single character separator between.
     * 
     * @return bool Validity of the given date
     * 
     * Examples:
     *  <code>
     *      isValidDate('22.22.2222', 'mm.dd.yyyy'); // returns false
     *      isValidDate('11/30/2008', 'mm/dd/yyyy'); // returns true
     *      isValidDate('30-01-2008', 'dd-mm-yyyy'); // returns true
     *      isValidDate('2008 01 30', 'yyyy mm dd'); // returns true
     *  </code>
     */
    public static function isValidDate(string $date, string $format = 'Y-m-d'):bool {
        return ($date === date($format, strtotime($date)));
    }
    
    /**
     * function isValidEmailAddress()
     * 
     * Checks if the email address has a valid format
     * 
     * @param string $email E-Mail address to validate
     * 
     * @return bool Validity of the given address
     */
    public static function isValidEmailAddress(string $email):bool {
        $oValidator = new EmailAddressValidator();
        $ret = $oValidator->check_email_address($email);
        unset($oValidator);
        return $ret;
    }

    /**
     * function isValidFloat()
     * 
     * Checks if the value is numeric and currency
     * 
     * @param string $number Value to check
     * @return bool Is the value a valid float
     */
    public static function isValidFloat(string $number):bool {
        $ret = false;
        $match = array();
        if (preg_match("/^[-+]?\b\d*(?:\.\d*)?\b/", $number, $match)) {
            $ret = ($match[0] == $number);
        }
        return $ret;
    }
    
    /**
     * function isValidIBAN()
     * 
     * Checks if the IBAN is valid
     * 
     * @param string $iban IBAN to validate
     * 
     * @return bool Validity of the given IBAN
     */
    public static function isValidIBAN(string $iban):bool {
        $iban   = strtoupper(str_replace(' ', '', $iban));
        $BBAN   = substr($iban, 4);
        $CTY    = (ord(substr($iban, 0, 1)) - 55) . (ord(substr($iban, 1, 1)) - 55);
        $CHK    = substr($iban, 2, 2);
        $check  = ($BBAN . $CTY . $CHK);
        $len    = 0;
        $run    = '';
        while (strlen($check)) {
            $len = (9 - strlen($run));
            $run = (intval($run . substr($check, 0, $len)) % 97);
            $check = substr($check, $len);
        }
        return ($run == 1);
    }

    /**
     * function isValidInteger()
     * 
     * Checks if the value is numeric and currency
     * 
     * @param string $number Value to check
     * @return bool Is the value a valid float
     */
    public static function isValidInteger(string $number):bool {
        $ret = false;
        $match = array();
        if (preg_match("/^[-+]?\b\d*\b/", $number, $match)) {
            $ret = ($match[0] == $number);
        }
        return $ret;
    }

    /**
     * function isValidIpAddress()
     * 
     * Checks if the value is numeric and currency
     * 
     * @param string $address Value to check
     * @return bool Is the value a valid float
     */
    public static function isValidIpAddress(string $address):bool {
        $ret = false;
        $match = array();
        if (preg_match("/\b\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\b/", $address, $match)) {
            $ret = ($match[0] == $address);
        }
        return $ret;
    }
    
    /**
     * function isValidName()
     * 
     * Checks if the name has a valid format
     * 
     * @param string $name Name to check
     * @param string $countryCode Country code the check the name validity for
     * @return bool Validity of the given name
     */
    public static function isValidName(string $name, string $countryCode = ''):bool {
        if (strlen($countryCode)) {
            $loc = setLocale(LC_ALL, 0);
            setLocale(LC_ALL, $countryCode);
        }
        $regex = '/^\D+(\s\D+)+$/';
        $ret = preg_match($regex, $name);
        if (strlen($countryCode)) {
            setlocale(LC_ALL, $loc);
        }
        return $ret;
    }
    
    /**
     * function isValidNumber()
     * 
     * Checks if the number has a valid format
     * 
     * @param string $number number to check
     * 
     * @return bool Validity of the given number
     */
    public static function isValidNumber(string $number):bool {
        $iDot = strpos($number, '.');
        $iComma = strpos($number, ',');
        if (($iDot !== false) && ($iComma !== false) && ($iDot > $iComma)) {
            return is_numeric($number);
        }
        elseif (($iDot !== false) && ($iComma !== false) && ($iDot < $iComma)) {
            return is_numeric(str_replace(array('.', ','), array('', '.'), $number));
        }
        elseif ($iDot !== false) {
            return is_numeric($number);
        }
        elseif ($iComma !== false) {
            return is_numeric(str_replace(',', '.', $number));
        }
    }
    
    /**
     * function isValidPhoneNumber()
     * 
     * Checks if the phone number has a valid format
     * 
     * @param string $phone Phone number to check
     * 
     * @return bool Validity of the given phone number
     */
    public static function isValidPhoneNumber(string $phone):bool {
        #$regex = '/^(\+[0-9]{2,3}|0+[0-9]{2,5}).+[\d\s\/\(\)-]/';
        # Allowed are either +xx xx.. xx.. (xx..) or 0xx.. xx.. (xx..)
        # Separators can be any of /, (space), - (- not after +xx)
        $regex = '/^(\+[0-9]{2,3}[\d\s\/\(\)\ ].+[0-9]{2,5}|0[0-9]{2,5})[\d\s\/\(\)\-\ ].+([0-9]{2,10})+$/';
        $regex = '/^(\+[0-9]{2,3}[\d\s\/\(\)\ ].+[0-9]{2,5}|0[0-9]{2,5})[\d\s\/\(\)\-\ ].+[0-9]{2,10}$/';
        return preg_match($regex, $phone);
    }
    
    /**
     * function isValidTime()
     * 
     * Checks if the time has a valid format
     * 
     * @param string $time Time to check
     * 
     * @return bool Validity of the given number
     */
    public static function isValidTime(string $time):bool {
        $ret = false;
        $match = array();
        if (preg_match("/\b\d{1,2}\:\d{2}(?:\:\d{2})?\b/", $time, $match)) {
            $ret = ($match[0] == $time);
        }
        return $ret;
    }
    
    /**
     * function isValidUrl()
     * 
     * Checks if the url has a valid format
     * 
     * @param string $url Url to check
     * 
     * @return bool Validity of the given url
     */
    public static function isValidUrl(string $url):bool {
        $regex = '/^(?:(?:https?|ftp):\/\/|www\.)[-a-z0-9+&@#%?=~_|!:,;]{2,}(\.[-a-z0-9+&@#\/%=~_|]{2,}){2,}$/i';
        return preg_match($regex, $url);
    }
    
    /**
     * function isValidZipCode()
     * 
     * Checks if the zip code has a valid format
     * 
     * @param string $zipCode ZipCode to check
     * @param string $countryCode The country code [us|de|nl|...]
     * @return bool Validity of the given zip code
     */
    public static function isValidZipCode(string $zipCode, string $countryCode):bool {
        $regex = '';
        if (strlen($countryCode)) {
            switch (strtoupper($countryCode)) {
                case 'AD':
                    $regex = '^AD\d{3}$';
                    break;
                case 'AE':
                    # No postal codes
                    break;
                case 'AF':
                    $regex = '^([1-2]\d|3[0-4])\d{2}$';
                    break;
                case 'AG':
                    # No postal codes
                    break;
                case 'AI':
                    return ($zipCode == 'AI-2640'); # single code for all addresses
                    break;
                case 'AL':
                    $regex = '^\d{4}$';
                    break;
                case 'AM':
                    $regex = '^\d{4}(?:\d{2})?$';
                    break;
                case 'AO':
                    # No postal codes
                    break;
                case 'AQ':
                    return ($zipCode == 'BIQQ 1ZZ'); # single code for all addresses
                    break;
                case 'AR':
                    $regex = '^(\d{4}|[A-H]|[J-N]|[P-Z]\d{4}[A-Z]{3})$';
                    break;
                case 'AS':
                    $regex = '^\d{5}(?:\-\d{4})?$';
                    break;
                case 'AT':
                    $regex = '^[1-9]\d{3}$';
                    break;
                case 'AU':
                    $regex = '^\d{4}$';
                    break;
                case 'AW':
                    # No postal codes
                    break;
                case 'AX':
                    $regex = '^(?:AX\-)?22\d{3}$';
                    break;
                case 'AZ':
                    $regex = '^(AZ\s\d{4}|\d{6})$';
                    break;
                case 'BA':
                    $regex = '^\d{5}$';
                    break;
                case 'BB':
                    $regex = '^BB\d{5}$';
                    break;
                case 'BD':
                    $regex = '^\d{4}$';
                    break;
                case 'BE':
                    $regex = '^\d{4}$';
                    break;
                case 'BF':
                    # No postal codes
                    break;
                case 'BG':
                    $regex = '^\d{4}$';
                    break;
                case 'BH':
                    $regex = '^([1-9](0[1-9]|1\d)|1([0-1]\d{2}|2[0-1][0-6]))$';
                    break;
                case 'BI':
                    # No postal codes
                    break;
                case 'BJ':
                    # No postal codes
                    break;
                case 'BL':
                    return ($zipCode == '97133'); # single code for all addresses
                    break;
                case 'BM':
                    $regex = '^[A-Z]{2}\s([BG]X|\d{2})$';
                    break;
                case 'BN':
                    $regex = '^[A-Z]{2}\d{4}$';
                    break;
                case 'BO':
                    # No postal codes
                    break;
                case 'BQ':
                    # No postal codes
                    break;
                case 'BR':
                    $regex = '^\d{5}(?:\-\d{3})?$';
                    break;
                case 'BS':
                    # No postal codes
                    break;
                case 'BT':
                    $regex = '^\d{5}$';
                    break;
                case 'BW':
                    # No postal codes
                    break;
                case 'BY':
                    $regex = '^\d{6}$';
                    break;
                case 'BZ':
                    # No postal codes
                    break;
                case 'CA':
                    $regex = '^([A-C]|E|[G-H]|[J-N]|P|[R-Z])\d([A-C]|E|[G-H]|[J-N]|P|[R-Z])\s\d([A-C]|E|[G-H]|[J-N]|P|[R-Z])\d$';
                    break;
                case 'CC':
                    $regex = '^\d{4}$';
                    break;
                case 'CD':
                    # No postal codes
                    break;
                case 'CF':
                    # No postal codes
                    break;
                case 'CG':
                    # No postal codes
                    break;
                case 'CH':
                    $regex = '^\d{4}$';
                    break;
                case 'CI':
                    # No postal codes
                    break;
                case 'CK':
                    # No postal codes
                    break;
                case 'CL':
                    $regex = '^\d{3}(?:\-)?\d{4}$';
                    break;
                case 'CM':
                    # No postal codes
                    break;
                case 'CN':
                    $regex = '^\d{6}$';
                    break;
                case 'CO':
                    $regex = '^([0-2]\d|3[0-2])\d{4}$';
                    break;
                case 'CR':
                    $regex = '^(\d{4}|\d{5}(?:\-\d{4})?)$';
                    break;
                case 'CU':
                    $regex = '^\d{5}$';
                    break;
                case 'CV':
                    $regex = '^\d{4}$';
                    break;
                case 'CW':
                    # No postal codes
                    break;
                case 'CX':
                    $regex = '^\d{4}(?:\d)?$';
                    break;
                case 'CZ':
                    $regex = '^[1-7]\d{2}\s\d{2}$';
                    break;
                case 'DE':
                    $regex = '^\d{5}$';
                    break;
                case 'DJ':
                    # No postal codes
                    break;
                case 'DK':
                    $regex = '^\d{5}$';
                    break;
                case 'DM':
                    # No postal codes
                    break;
                case 'DZ':
                    $regex = '^([0-3]\d|4[0-8])\d{3}$';
                    break;
                case 'EC':
                    $regex = '^\d{6}$';
                    break;
                case 'EE':
                    $regex = '^\d{5}$';
                    break;
                case 'ER':
                    # No postal codes
                    break;
                case 'ES':
                    $regex = '^([0-4]\d|5[0-2])\d{3}$';
                    break;
                case 'ET':
                    $regex = '^\d{4}$';
                    break;
                case 'FI':
                    $regex = '^\d{4}[0-1]$';
                    break;
                case 'FJ':
                    # No postal codes
                    break;
                case 'FK':
                    return ($zipCode == 'FIQQ 1ZZ'); # single code for all addresses
                    break;
                case 'FM':
                    $regex = '^\d{5}(?:\-\d{4})?$';
                    break;
                case 'FO':
                    $regex = '^FO\-\d{3}$';
                    break;
                case 'FR':
                    $regex = '^\d{5}$';
                    break;
                case 'GA':
                    # No postal codes
                    break;
                case 'GB':
                    $regex = '^[A-Z]{1,2}\d[\dA-Z]?\s?\d[A-Z]{2}$';
                    break;
                case 'GD':
                    # No postal codes
                    break;
                case 'GE':
                    $regex = '^\d{4}$';
                    break;
                case 'GF':
                    $regex = '^973([0-8]\d|90)$';
                    break;
                case 'GG':
                    $regex = '^GY\d{1,2}\s\d[A-Z]{2}$';
                    break;
                case 'GH':
                    $regex = '^[A-Z]{2}\-\d{4}(?:\-\d{4})?$';
                    break;
                case 'GI':
                    return ($zipCode == 'GX11 1AA'); # single code for all addresses
                    break;
                case 'GL':
                    $regex = '^\d{4}$';
                    break;
                case 'GM':
                    # No postal codes
                    break;
                case 'GN':
                    $regex = '^\d{3}$';
                    break;
                case 'GP':
                    $regex = '^971([0-8]\d|90)$';
                    break;
                case 'GQ':
                    # No postal codes
                    break;
                case 'GR':
                    $regex = '^\d{3}\s\d{2}$';
                    break;
                case 'GS':
                    return ($zipCode == 'SIQQ 1ZZ'); # single code for all addresses
                    break;
                case 'GT':
                    $regex = '^\d{5}$';
                    break;
                case 'GU':
                    $regex = '^969([1-2]\d|3[1-2])$';
                    break;
                case 'GW':
                    $regex = '^\d{4}$';
                    break;
                case 'GY':
                    # No postal codes
                    break;
                case 'HK':
                    # No postal codes
                    break;
                case 'HM':
                    # No postal codes
                    break;
                case 'HN':
                    $regex = '^([A-Z]{2}\d{4}|\d{5})$';
                    break;
                case 'HR':
                    $regex = '^\d{5}$';
                    break;
                case 'HT':
                    $regex = '^\d{4}$';
                    break;
                case 'HU':
                    $regex = '^\d{4}$';
                    break;
                case 'ID':
                    $regex = '^\d{5}$';
                    break;
                case 'IE':
                    $regex = '^(A|[C-F]|H|K|N|P|R|T|[V-X])\d{2}\s\w{4}$';
                    break;
                case 'IL':
                    $regex = '^\d{5}(?:\d{2})?$';
                    break;
                case 'IM':
                    $regex = '^IM\d(?:\d)?\s\d[A-Z]{2}$';
                    break;
                case 'IN':
                    $regex = '^\d{3}(?:\s)?\d{3}$';
                    break;
                case 'IO':
                    return ($zipCode == 'BBND 1ZZ'); # single code for all addresses
                    break;
                case 'IQ':
                    $regex = '^\d{5}$';
                    break;
                case 'IR':
                    $regex = '^\d{5}\-\d{5}$';
                    break;
                case 'IS':
                    $regex = '^\d{3}$';
                    break;
                case 'IT':
                    $regex = '^\d{5}$';
                    break;
                case 'JE':
                    $regex = '^JE\d(?:\d)?\s\d[A-Z]{2}$';
                    break;
                case 'JM':
                    $regex = '^\d{1,2}$';
                    break;
                case 'JO':
                    $regex = '^\d{5}$';
                    break;
                case 'JP':
                    $regex = '^\d{3}\-\d{4}$';
                    break;
                case 'KE':
                    $regex = '^\d{5}$';
                    break;
                case 'KG':
                    $regex = '^\d{6}$';
                    break;
                case 'KH':
                    return ($zipCode == '120000'); # single code for all addresses
                    break;
                case 'KI':
                    # No postal codes
                    break;
                case 'KM':
                    # No postal codes
                    break;
                case 'KN':
                    $regex = '^\KN\d{4}$';
                    break;
                case 'KP':
                    # No postal codes
                    break;
                case 'KR':
                    $regex = '^\d{5}$';
                    break;
                case 'KW':
                    $regex = '^\d{5}$';
                    break;
                case 'KY':
                    $regex = '^\KY\d\-\d{4}$';
                    break;
                case 'KZ':
                    $regex = '^\d{6}$';
                    break;
                case 'LA':
                    $regex = '^\d{5}$';
                    break;
                case 'LB':
                    $regex = '^(\d{5}|\d{4}\s\d{4})$';
                    break;
                case 'LC':
                    $regex = '^LC\d{2}\s\d{3}$';
                    break;
                case 'LI':
                    $regex = '^94(8[5-9]|9[0-8])$';
                    break;
                case 'LK':
                    $regex = '^\d{5}$';
                    break;
                case 'LR':
                    $regex = '^\d{2}(?:\d{2})?$';
                    break;
                case 'LS':
                    $regex = '^\d{3}$';
                    break;
                case 'LT':
                    $regex = '^LT\-\d{5}$';
                    break;
                case 'LU':
                    $regex = '^\d{4}$';
                    break;
                case 'LV':
                    $regex = '^LV\-\d{4}$';
                    break;
                case 'LY':
                    # No postal codes
                    break;
                case 'MA':
                    $regex = '^\d{5}$';
                    break;
                case 'MC':
                    $regex = '^\980d{2}$';
                    break;
                case 'MD':
                    $regex = '^\MD(?:\-)?d{4}$';
                    break;
                case 'ME':
                    $regex = '^\d{4}$';
                    break;
                case 'MF':
                    return ($zipCode == '97150'); # single code for all addresses
                    break;
                case 'MG':
                    $regex = '^\d{3}$';
                    break;
                case 'MH':
                    $regex = '^\969(6\d|70)(?:\-\d{4})?$';
                    break;
                case 'MK':
                    $regex = '^\d{4}$';
                    break;
                case 'ML':
                    # No postal codes
                    break;
                case 'MM':
                    $regex = '^\d{5}$';
                    break;
                case 'MN':
                    $regex = '^\d{5}$';
                    break;
                case 'MO':
                    # No postal codes
                    break;
                case 'MP':
                    $regex = '^\9695[0-2](?:\-\d{4})?$';
                    break;
                case 'MQ':
                    $regex = '^\972([0-8]\d|90)$';
                    break;
                case 'MR':
                    # No postal codes
                    break;
                case 'MS':
                    $regex = '^MSR\s1(1[1-9]\d|2\d{2}|3[0-4]\d|350)$';
                    break;
                case 'MT':
                    $regex = '^[A-Z]{3}\s\d{4}$';
                    break;
                case 'MU':
                    $regex = '^\d{5}$';
                    break;
                case 'MV':
                    $regex = '^\d{5}$';
                    break;
                case 'MW':
                    $regex = '^\d{6}$';
                    break;
                case 'MX':
                    $regex = '^\d{5}$';
                    break;
                case 'MY':
                    $regex = '^\d{5}$';
                    break;
                case 'MZ':
                    $regex = '^\d{4}$';
                    break;
                case 'NA':
                    $regex = '^\d{5}$';
                    break;
                case 'NC':
                    $regex = '^988([0-8]\d|90)$';
                    break;
                case 'NE':
                    $regex = '^\d{4}$';
                    break;
                case 'NF':
                    $regex = '^\d{4}$';
                    break;
                case 'NG':
                    $regex = '^\d{5}$';
                    break;
                case 'NI':
                    $regex = '^\d{4}$';
                    break;
                case 'NL':
                    $regex = '^\d{4}(?:\s{1})?[A-Z]{2}$';
                    break;
                case 'NO':
                    $regex = '^\d{4}$';
                    break;
                case 'NP':
                    $regex = '^\d{5}$';
                    break;
                case 'NR':
                    # No postal codes
                    break;
                case 'NU':
                    # No postal codes
                    break;
                case 'NZ':
                    $regex = '^\d{4}$';
                    break;
                case 'OM':
                    $regex = '^\d{3}$';
                    break;
                case 'PA':
                    $regex = '^\d{4}$';
                    break;
                case 'PE':
                    $regex = '^(\d{5}|PE\s\d{4})$';
                    break;
                case 'PF':
                    $regex = '^987([0-8]\d|90)$';
                    break;
                case 'PG':
                    $regex = '^\d{3}$';
                    break;
                case 'PH':
                    $regex = '^\d{4}$';
                    break;
                case 'PK':
                    $regex = '^\d{5}$';
                    break;
                case 'PG':
                    $regex = '^\d{2}\-\d{3}$';
                    break;
                case 'PM':
                    return ($zipCode == '97500'); # single code for all addresses
                    break;
                case 'PN':
                    return ($zipCode == 'PCRN 1ZZ'); # single code for all addresses
                    break;
                case 'PR':
                    $regex = '^00[6|7|9]\d{2}(?:\-\d{4})?$';
                    break;
                case 'PS':
                    $regex = '^\d{3}$';
                    break;
                case 'PT':
                    $regex = '^\d{4}(?:\-\d{3})?$';
                    break;
                case 'PW':
                    $regex = '^969[39|40](?:\-\d{4})?$';
                    break;
                case 'PY':
                    $regex = '^\d{4}$';
                    break;
                case 'QA':
                    # No postal codes
                    break;
                case 'RE':
                    $regex = '^974([0-8]\d|90)$';
                    break;
                case 'RO':
                    $regex = '^\d{6}$';
                    break;
                case 'RS':
                    $regex = '^\d{5}$';
                    break;
                case 'RU':
                    $regex = '^\d{6}$';
                    break;
                case 'RW':
                    # No postal codes
                    break;
                case 'SA':
                    $regex = '^\d{5}(?:\-\d{4})?$';
                    break;
                case 'SB':
                    # No postal codes
                    break;
                case 'SC':
                    # No postal codes
                    break;
                case 'SD':
                    $regex = '^\d{5}$';
                    break;
                case 'SE':
                    $regex = '^\d{3}\s\d{2}$';
                    break;
                case 'SG':
                    $regex = '^\d{2}(?:\d{2})?(?:\d{2})?$';
                    break;
                case 'SH':
                    $regex = '^[STHL|ASCN|TDCU]\s1ZZ$';
                    break;
                case 'SI':
                    $regex = '^(?:SI\s)?\d{4}$';
                    break;
                case 'SJ':
                    $regex = '^\d{4}$';
                    break;
                case 'SK':
                    $regex = '^\d{3}\s\d{2}$';
                    break;
                case 'SL':
                    # No postal codes
                    break;
                case 'SM':
                    $regex = '^4789\d$';
                    break;
                case 'SN':
                    $regex = '^\d{5}$';
                    break;
                case 'SO':
                    $regex = '^[A-Z]{2}\s\d{5}$';
                    break;
                case 'SR':
                    # No postal codes
                    break;
                case 'SS':
                    # No postal codes
                    break;
                case 'ST':
                    # No postal codes
                    break;
                case 'SV':
                    $regex = '^\d{4}$';
                    break;
                case 'SX':
                    # No postal codes
                    break;
                case 'SY':
                    $regex = '^\d{4}$';
                    break;
                case 'SZ':
                    $regex = '^[H|M|S|L]\d{3}$';
                    break;
                case 'TC':
                    return ($zipCode == 'TKCA 1ZZ'); # single code for all addresses
                    break;
                case 'TD':
                    # No postal codes
                    break;
                case 'TF':
                    # No postal codes
                    break;
                case 'TG':
                    # No postal codes
                    break;
                case 'TH':
                    $regex = '^\d{5}$';
                    break;
                case 'TJ':
                    $regex = '^\d{6}$';
                    break;
                case 'TK':
                    # No postal codes
                    break;
                case 'TL':
                    # No postal codes
                    break;
                case 'TM':
                    $regex = '^\d{6}$';
                    break;
                case 'TN':
                    $regex = '^\d{4}$';
                    break;
                case 'TO':
                    # No postal codes
                    break;
                case 'TR':
                    $regex = '^\d{5}$';
                    break;
                case 'TT':
                    $regex = '^\d{6}$';
                    break;
                case 'TV':
                    # No postal codes
                    break;
                case 'TW':
                    $regex = '^\d{3}(?:\-\d{2,3})?$';
                    break;
                case 'TZ':
                    $regex = '^\d{5}$';
                    break;
                case 'UA':
                    $regex = '^\d{5}$';
                    break;
                case 'UG':
                    # No postal codes
                    break;
                case 'UM':
                    return ($zipCode == '96898'); # single code for all addresses
                    break;
                case 'US':
                    $regex = '^\d{5}(?:\-\d{4})?$';
                    break;
                case 'UY':
                    $regex = '^\d{5}$';
                    break;
                case 'UZ':
                    $regex = '^\d{6}$';
                    break;
                case 'VA':
                    return ($zipCode == '00120'); # single code for all addresses
                    break;
                case 'VC':
                    $regex = '^VC\d{4}$';
                    break;
                case 'VE':
                    $regex = '^\d{4}(?:\-[A-Z])?$';
                    break;
                case 'VG':
                    $regex = '^VG11([1-5]\d|60)$';
                    break;
                case 'VI':
                    $regex = '^008(0[1-9]|[1-4]\d|5[0-1])$';
                    break;
                case 'VN':
                    $regex = '^\d{5}$';
                    break;
                case 'VU':
                    # No postal codes
                    break;
                case 'WF':
                    $regex = '^986([0-8]\d|90)$';
                    break;
                case 'WS':
                    $regex = '^WS\d{4}$';
                    break;
                case 'XK':
                    $regex = '^\d{5}$';
                    break;
                case 'YE':
                    # No postal codes
                    break;
                case 'YT':
                    $regex = '^976([0-8]\d|90)$';
                    break;
                case 'ZA':
                    $regex = '^\d{5}$';
                    break;
                case 'ZM':
                    $regex = '^\d{5}$';
                    break;
                case 'ZW':
                    # No postal codes
                    break;
            }
            if (strlen($regex)) {
                $matches = array();
                if (preg_match($regex, $zipCode, $matches)) {
                    return ($matches[0] == $zipCode);
                }
            }
            return false;
        }
        else {
            return true; /* Without a country code, we can't test the zip code */
        }
    }
}
?>