<?php
/**
 * Project: 
 * DZIRCMS Content Management System
 * 
 * Description: 
 * HTML5 form, elements, and checks
 * 
 * Requirements: 
 * @con_php_req 7.4
 * 
 * 
 * @package     DZIRCMS Frontend classes
 * @version     1.0.0
 * @author      René Mansveld
 * @copyright   Spider IT Deutschland 
 * @license     https://www.dzir.org/cms/license/licence.txt
 * @link        https://www.spider-it.de
 * @link        https://www.dzir.org/cms/
 * @since       file available since DZIRCMS release 2.2.0
 * 
 * {@internal
 *  created:    2022-08-31
 * }}
 */

abstract class NumericFieldTypes {
    const currency = 'currency';
    const date = 'date';
    const float = 'float';
    const integer = 'integer';
    const time = 'time';
}

abstract class LabelPositions {
    const before = 'before';
    const after = 'after';
}

abstract class ButtonTypes {
    const button = 'button';
    const reset = 'reset';
    const submit = 'submit';
}

class html5form {

    /**
     * @var array $aForm
     * 
     * Holds information about the form itself
     */
    protected array $aForm = array();

    /**
     * @var array $aHiddens
     * 
     * Holds information about the hidden fields in the form
     */
    protected array $aHiddens = array();

    /**
     * @var array $aFields
     * 
     * Holds information about the fields in the form
     */
    protected array $aFields = array();

    /**
     * @var array $aButtons
     * 
     * Holds information about the buttons in the form
     */
    protected array $aButtons = array();

    /**
     * Constructor
     * 
     * @param string $formUrl The url for the form to send to
     * @param string $formMethod The method for the form to send the data (get|post)
     * @param string $formName The name of the form
     * @param string $formId The html id for the form
     * @param string $formClass The CSS class name for the form
     */
    public function __construct(string $formUrl = '', string $formMethod = 'post', string $formName = '', string $formId = '', string $formClass = '') {
        $this->aForm = array(
            'url' => $formUrl,
            'method' => $formMethod,
            'name' => $formName,
            'id' => $formId,
            'class' => $formClass
        );
    }

    /**
     * function setFormUrl()
     * 
     * Sets the form url
     * 
     * @param string $url The url of the form
     */
    public function setFormUrl(string $url) {
        $this->aForm['url'] = $url;
    }

    /**
     * function setFormMethod()
     * 
     * Sets the form method
     * 
     * @param string $method The method of the form
     */
    public function setFormMethod(string $method) {
        $this->aForm['method'] = $method;
    }

    /**
     * function setFormName()
     * 
     * Sets the form name
     * 
     * @param string $name The name of the form
     */
    public function setFormName(string $name) {
        $this->aForm['name'] = $name;
    }

    /**
     * function setFormId()
     * 
     * Sets the form id
     * 
     * @param string $id The id of the form
     */
    public function setFormId(string $id) {
        $this->aForm['id'] = $id;
    }

    /**
     * function setFormClass()
     * 
     * Sets the form class
     * 
     * @param string $class The class of the form
     */
    public function setFormClass(string $class) {
        $this->aForm['class'] = $class;
    }

    /**
     * function addBicField()
     * 
     * Adds a label and field for bank account BIC values
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addBicField(string $label, string $input, $value = '', bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'bic',
            'value' => $value,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addCheckbox()
     * 
     * Ads a label and checkbox field with additional descriptive label
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param bool $checked Is this element preselected
     * @param bool $required Must this field be non empty
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $descriptiveLabel A descriptive label on the other side of the checkbox
     * @param string $descriptiveLabelClass The class name of the descriptive label
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addCheckbox(string $label, string $input, $value = '1', bool $checked = false, bool $required = false, string $labelPosition = LabelPositions::before, string $descriptiveLabel = '', string $descriptiveLabelClass = '', string $className = '', $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'checkbox',
            'value' => $value,
            'checked' => $checked,
            'required' => $required,
            'labelPosition' => $labelPosition,
            'descriptiveLabel' => $descriptiveLabel,
            'descriptiveLabelClass' => $descriptiveLabelClass,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addCheckboxGroup()
     * 
     * Ads a label and checkbox fields with additional descriptive labels
     * 
     * @param string $label The label text to show
     * @param array $values Array of checkbox data
     *                  array(
     *                      array(
     *                          <input> The name of the input
     *                          <value> The value of this checkbox
     *                          <descriptiveLabel> A descriptive label on the side of the checkbox
     *                          <checked> Is this checkbox checked
     *                          <required> Must one radio be selected
     *                      ), ...
     *                  )
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $descriptiveLabelClass The class name of the descriptive label
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addCheckboxGroup(string $label, array $values, string $labelPosition = LabelPositions::before, string $descriptiveLabelClass = '', string $className = '', $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'type' => 'checkboxGroup',
            'values' => $values,
            'labelPosition' => $labelPosition,
            'descriptiveLabelClass' => $descriptiveLabelClass,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addColorField()
     * 
     * Adds a label and field for bank account BIC values
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param bool $required Must this field be non empty
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addColorField(string $label, string $input, $value = '', bool $required = false, array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'color',
            'value' => $value,
            'required' => $required,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addCreditCardField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addCreditCardField(string $label, string $input, $value = '', bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'creditcard',
            'value' => $value,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addDateField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param string|null $min The minimum allowed value
     * @param string|null $max The maximum allowed value
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addDateField(string $label, string $input, $value = '', $min = null, $max = null, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'date',
            'value' => ((strlen($value)) ? date('Y-m-d', strtotime($value)) : $value),
            'min' => (($min !== null) ? date('Y-m-d', strtotime($min)) : $min),
            'max' => (($max !== null) ? date('Y-m-d', strtotime($max)) : $max),
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addEmailField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addEmailField(string $label, string $input, $value = '', bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'email',
            'value' => $value,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addFileField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param array $filters The file type filters
     *              Example: array('*.png', '*.jpg', '*.jpeg', '*.gif')
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addFileField(string $label, string $input, $value = '', array $filters = array(), bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'file',
            'value' => $value,
            'filters' => $filters,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addHiddenField()
     * 
     * Adds a hidden field to the form
     * 
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     */
    public function addHiddenField(string $input, $value = '') {
        $this->aHiddens[] = array(
            'input' => $input,
            'value' => $value,
        );
    }

    /**
     * function addIbanField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addIbanField(string $label, string $input, $value = '', bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'iban',
            'value' => $value,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addIpField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addIpField(string $label, string $input, $value = '', bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'ip',
            'value' => $value,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addNameField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param int|null $min The minimum allowed length
     * @param int|null $max The maximum allowed length
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     * @param string $countryCode The 2 (ISO 3366-1 alpha-2, https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) character country code
     */
    public function addNameField(string $label, string $input, $value = '', $min = null, $max = null, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '', string $countryCode = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'name',
            'value' => $value,
            'min' => $min,
            'max' => $max,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss,
            'countrycode' => $countryCode
        );
    }

    /**
     * function addNumericField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param NumericFieldTypes $type The type name of the input (integer|number|currency)
     * @param float|null $value The preset value of the input
     * @param float|null $min The minimum allowed value
     * @param float|null $max The maximum allowed value
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addNumericField(string $label, string $input, string $type = NumericFieldTypes::integer, $value = null, $min = null, $max = null, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => $type,
            'value' => ((($value !== null) && ($type == NumericFieldTypes::integer)) ? (int) $value : $value),
            'min' => ((($min !== null) && ($type == NumericFieldTypes::integer)) ? (int) $min : $min),
            'max' => ((($max !== null) && ($type == NumericFieldTypes::integer)) ? (int) $max : $max),
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addPasswordField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param int|null $min The minimum allowed length
     * @param int|null $max The maximum allowed length
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addPasswordField(string $label, string $input, $value = '', $min = null, $max = null, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'password',
            'value' => $value,
            'min' => $min,
            'max' => $max,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addRadio()
     * 
     * Ads a label and radio field with additional descriptive label
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The value of the input
     * @param bool $checked Is this element preselected
     * @param bool $required Must this field be non empty
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $descriptiveLabel A descriptive label on the other side of the radio
     * @param string $descriptiveLabelClass The class name of the descriptive label
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addRadio(string $label, string $input, $value = '1', bool $checked = false, bool $required = false, string $labelPosition = LabelPositions::before, string $descriptiveLabel = '', string $descriptiveLabelClass = '', string $className = '', $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'radio',
            'value' => $value,
            'checked' => $checked,
            'required' => $required,
            'labelPosition' => $labelPosition,
            'descriptiveLabel' => $descriptiveLabel,
            'descriptiveLabelClass' => $descriptiveLabelClass,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addRadioGroup()
     * 
     * Ads a label and radio fields with additional descriptive labels
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param array $values Array of radio button data
     *                  array(
     *                      array(
     *                          <value> The value of this radio
     *                          <descriptiveLabel> A descriptive label on the side of the radio
     *                          <checked> Is this radio checked (only one in the group)
     *                      ), ...
     *                  )
     * @param bool $required Must one radio be selected
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $descriptiveLabelClass The class name of the descriptive label
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addRadioGroup(string $label, string $input, array $values, bool $required = false, string $labelPosition = LabelPositions::before, string $descriptiveLabelClass = '', string $className = '', $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'radioGroup',
            'values' => $values,
            'required' => $required,
            'labelPosition' => $labelPosition,
            'descriptiveLabelClass' => $descriptiveLabelClass,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addRange()
     * 
     * Adds a label and range input
     * 
     * @param string  $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param int|null $min The minimum allowed value
     * @param int|null $max The maximum allowed value
     * @param int $step The step width
     * @param bool $required Must this field be non empty
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $descriptiveLabelClass The class name of the descriptive label
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addRange(string $label, string $input, $value = '', $min = null, $max = null, $step = 1, bool $required = false, array $restrictions = array(), string $labelPosition = LabelPositions::before, $descriptiveLabelClass = '', string $className = '', string $additionalCss = '') {
        if ((strlen($value)) && ($min !== null)) {
            #echo '<pre>((' . $value . ' % ' . $step . ') != (' . $min . ' % ' . $step . ') ==&gt; (' . ($value % $step) . ' != ' . ($min % $step) . ') => ' . (($value % $step) != ($min % $step) ? 'true' : 'false') . '</pre>';
            if (($value % $step) != ($min % $step)) {

                # Round the $value up/down to the nearest step
                #echo '<pre>$value = ((round((' . $value . ' + (abs(' . $min . ') % ' . $step . ')) / ' . $step . ') * ' . $step . ') + (' . $min . ' % ' . $step . '))</pre>';
                $value = ((round(($value + (abs($min) % $step)) / $step) * $step) + ($min % $step));
                #echo '<pre>$value = ' . $value . '</pre>';
            }
        }
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'range',
            'value' => $value,
            'min' => $min,
            'max' => $max,
            'step' => $step,
            'required' => $required,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'descriptiveLabel' => $value,
            'descriptiveLabelClass' => $descriptiveLabelClass,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addSelect()
     * 
     * Adds a label and select field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param array $options The options for the select (name [option|optgroup], value (optional), text)
     * @param string|null $value The preset value of the input
     * @param bool $multiple Is this a multi-select
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addSelect(string $label, string $input, array $options, $value = '', bool $multiple = false, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'select',
            'options' => $options,
            'value' => $value,
            'multiple' => $multiple,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addTelField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param int|null $min The minimum allowed length
     * @param int|null $max The maximum allowed length
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addTelField(string $label, string $input, $value = '', $min = null, $max = null, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'tel',
            'value' => $value,
            'min' => $min,
            'max' => $max,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addTextField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param int|null $min The minimum allowed length
     * @param int|null $max The maximum allowed length
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addTextField(string $label, string $input, $value = '', $min = null, $max = null, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'text',
            'value' => $value,
            'min' => $min,
            'max' => $max,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addTextarea()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param int|null $min The minimum allowed length
     * @param int|null $max The maximum allowed length
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addTextarea(string $label, string $input, $value = '', $min = null, $max = null, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'textarea',
            'value' => $value,
            'min' => $min,
            'max' => $max,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addTimeField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param string|null $min The minimum allowed value
     * @param string|null $max The maximum allowed value
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addTimeField(string $label, string $input, $value = '', $min = null, $max = null, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'time',
            'value' => ((strlen($value)) ? date('H:i:s', strtotime($value)) : $value),
            'min' => (($min !== null) ? date('H:i:s', strtotime($min)) : $min),
            'max' => (($max !== null) ? date('H:i:s', strtotime($max)) : $max),
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addUrlField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param int|null $min The minimum allowed length
     * @param int|null $max The maximum allowed length
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     */
    public function addUrlField(string $label, string $input, $value = '', $min = null, $max = null, bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'url',
            'value' => $value,
            'min' => $min,
            'max' => $max,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss
        );
    }

    /**
     * function addZipcodeField()
     * 
     * Adds a label and field to the form
     * 
     * @param string $label The label text to show
     * @param string $input The name of the input
     * @param string|null $value The preset value of the input
     * @param bool $required Must this field be non empty
     * @param string $placeholder Placeholder text
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     * @param string $className The name of the additional class
     * @param string $additionalCss Additional CSS to the input
     * @param string $countryCode The 2 (ISO 3366-1 alpha-2, https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2) character country code
     */
    public function addZipcodeField(string $label, string $input, $value = '', bool $required = false, string $placeholder = '', array $restrictions = array(), string $labelPosition = LabelPositions::before, string $className = '', string $additionalCss = '', string $countryCode = '') {
        $this->aFields[] = array(
            'label' => $label,
            'input' => $input,
            'type' => 'zipcode',
            'value' => $value,
            'required' => $required,
            'placeholder' => $placeholder,
            'restrictions' => $restrictions,
            'labelPosition' => $labelPosition,
            'class' => $className,
            'additionalCss' => $additionalCss,
            'countrycode' => $countryCode
        );
    }

    /**
     * function addCode()
     * 
     * Adds html code in between of the fields defined before and after the code
     * 
     * @param string $code The html code to insert
     */
    public function addCode(string $code) {
        $this->aFields[] = array(
            'label' => '',
            'type' => 'code',
            'value' => $code
        );
    }

    /**
     * function addButton()
     * 
     * Adds a button to the form
     * 
     * @param string|null $value The text to show on the button
     * @param ButtonTypes $type The type of the button (submit|reset|button)
     * @param string $name The name of the button
     * @param string $id The id name of the button
     * @param string $class The class name of the button
     * @param string $additionalCss Additional CSS to the input
     * @param string $label If set, adds a label for the buttons (only use on the first|last button)
     * @param string $labelClass The name of the class to add to the label
     * @param LabelPositions $labelPosition The position of the label relative to the input (before|after)
     */
    public function addButton(string $value, string $type = ButtonTypes::submit, string $name = '', string $id = '', string $class = '', string $additionalCss = '', string $label = '', string $labelClass = '', string $labelPosition = LabelPositions::before) {
        $this->aButtons[] = array(
            'value' => $value,
            'type' => $type,
            'name' => $name,
            'id' => $id,
            'class' => $class,
            'additionalCss' => $additionalCss,
            'label' => $label,
            'labelClass' => $labelClass,
            'labelPosition' => $labelPosition
        );
    }

    /**
     * function display()
     * 
     * Compiles and displays the form
     */
    public function display() {
        $aHiddens = array();
        $aFields = array();
        $aScripts = array();

        # Main script
        $aScripts['preload'] = ((strlen($this->aForm['id'])) ? $this->aForm['id'] : 'Form-0');
        $aScripts['form'] = ((strlen($this->aForm['id'])) ? $this->aForm['id'] : 'Form-0');

        # Hidden fields
        if (count($this->aHiddens)) {
            foreach ($this->aHiddens as $aHidden) {
                $aHiddens[] = '<input type="hidden" name="' . $aHidden['input'] . '" value="' . $aHidden['value'] . '">';
            }
        }

        # Visible fields
        $i = 0;
        foreach ($this->aFields as $aField) {
            $i ++;
            if ($aField['type'] == 'checkboxGroup') {
                $iID = 'F-' . str_replace(array('[', ']'), '', $aField['values'][0][0]) . '-' . $i;
            }
            else {
                $iID = 'F-' . str_replace(array('[', ']'), '', $aField['input']) . '-' . $i;
            }
            if ($aField['type'] != 'code') {
                $aFields[] = '<div class="row">';
            }
            if (($aField['labelPosition'] == 'before') && (strlen($aField['label']))) {
                $aFields[] = '<label' . ((strlen($iID)) ? ' for="' . $iID . '"' : '') . '>' . $aField['label'] . '</label>';
            }
            $aRestrictions = array();
            foreach ($aField['restrictions'] as $aRestriction) {
                $aRestrictions[] = $aRestriction['function'] . '~' . $aRestriction['operator'] . '~' . $aRestriction['compare'];
            }
            switch ($aField['type']) {
                case 'select':
                    $aFields[] = '<select id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . (($aField['multiple']) ? ' multiple' : '') . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>';
                    foreach ($aField['options'] as $aOption) {
                        if ($aOption['name'] == 'optgroup') {
                            $aFields[] = '<optgroup label="' . $aOption['text'] . '"></optgroup>';
                        }
                        else {
                            $aFields[] = '<option value="' . ((strlen($aOption['value'])) ? $aOption['value'] : $aOption['text']) . '"' . ((((strlen($aOption['value'])) ? $aOption['value'] : $aOption['text']) == $aField['value']) ? ' selected' : '') . '>' . $aOption['text'] . '</option>';
                        }
                    }
                    $aFields[] = '</select>';
                    break;
                case 'textarea':
                    $aFields[] = '<textarea id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . (($aField['min'] !== null) ? ' min="' . $aField['min'] . '"' : '') . (($aField['max'] !== null) ? ' max="' . $aField['max'] . '"' : '') . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>' . $aField['value'] . '</textarea>';
                    break;
                case 'currency':
                case 'float':
                case 'integer':
                    $aFields[] = '<input type="number" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aField['value'] . '"' . (($aField['min'] !== null) ? ' min="' . $aField['min'] . '"' : '') . (($aField['max'] !== null) ? ' max="' . $aField['max'] . '"' : '') . (($aField['type'] == 'integer') ? ' step="1"' : (($aField['type'] == 'currency') ? ' step=".01"' : ' step="any"')) . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>';
                    break;
                case 'bic':
                case 'creditcard':
                case 'iban':
                case 'ip':
                case 'url':
                case 'zipcode':
                    $aFields[] = '<input type="text" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aField['value'] . '"' . (($aField['min'] !== null) ? ' min="' . $aField['min'] . '"' : '') . (($aField['max'] !== null) ? ' max="' . $aField['max'] . '"' : '') . (($aField['countrycode']) ? ' data-countrycode="' . $aField['countrycode'] . '"' : '') . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>';
                    break;
                case 'name':
                    $aFields[] = '<input type="text" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aField['value'] . '"' . (($aField['min'] !== null) ? ' min="' . $aField['min'] . '"' : '') . (($aField['max'] !== null) ? ' max="' . $aField['max'] . '"' : '') . (($aField['countrycode']) ? ' data-countrycode="' . $aField['countrycode'] . '"' : '') . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>';
                    break;
                case 'checkbox':
                case 'radio':
                    $a = array();
                    $a[] = '<label for="' . $iID . '"' . ((strlen($aField['descriptiveLabelClass'])) ? ' class="' . $aField['descriptiveLabelClass'] . '"' : '') . '>';
                    if (($aField['labelPosition'] == 'after') && (strlen($aField['descriptiveLabel']))) {
                        $a[] = $aField['descriptiveLabel'] . ' ';
                    }
                    $a[] = '<input type="' . $aField['type'] . '" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aField['value'] . '"' . (($aField['required']) ? ' required' : '') . (($aField['checked']) ? ' checked' : '') . '>';
                    if (($aField['labelPosition'] == 'before') && (strlen($aField['descriptiveLabel']))) {
                        $a[] = ' ' . $aField['descriptiveLabel'];
                    }
                    $a[] = '</label>';
                    $aFields[] = implode(' ', $a);
                    break;
                case 'checkboxGroup':
                    $a = array('<div id="' . $iID . '" class="checkboxgroup" style="display: inline-block;">');
                    foreach ($aField['values'] as $aValue) {
                        $i ++;
                        $iID = 'F-' . str_replace(array('[', ']'), '', $aValue[0]) . '-' . $i;
                        $a[] = '<label for="' . $iID . '"' . ((strlen($aField['descriptiveLabelClass'])) ? ' class="' . $aField['descriptiveLabelClass'] . '"' : '') . '>';
                        if (($aField['labelPosition'] == 'after') && (strlen($aValue[2]))) {
                            $a[] = $aValue[2] . ' ';
                        }
                        $a[] = '<input type="checkbox" id="' . $iID . '" name="' . $aValue[0] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="checkbox"' . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aValue[1] . '"' . (($aValue[4]) ? ' required' : '') . (($aValue[3]) ? ' checked' : '') . '>';
                        if (($aField['labelPosition'] == 'before') && (strlen($aValue[2]))) {
                            $a[] = ' ' . $aValue[2];
                        }
                        $a[] = '</label> &nbsp; ';
                    }
                    $aFields[] = substr(implode('', $a), 0, -8) . '</div>';;
                    break;
                case 'radioGroup':
                    $a = array('<div id="' . $iID . '" class="radiogroup" style="display: inline; padding: 1px 2px;">');
                    foreach ($aField['values'] as $aValue) {
                        $i ++;
                        $iID = 'F-' . str_replace(array('[', ']'), '', $aField['input']) . '-' . $i;
                        $a[] = '<label for="' . $iID . '"' . ((strlen($aField['descriptiveLabelClass'])) ? ' class="' . $aField['descriptiveLabelClass'] . '"' : '') . '>';
                        if (($aField['labelPosition'] == 'after') && (strlen($aValue[1]))) {
                            $a[] = $aValue[1] . ' ';
                        }
                        $a[] = '<input type="radio" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="radio"' . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aValue[0] . '"' . (($aField['required']) ? ' required' : '') . (($aValue[2]) ? ' checked' : '') . '>';
                        if (($aField['labelPosition'] == 'before') && (strlen($aValue[1]))) {
                            $a[] = ' ' . $aValue[1];
                        }
                        $a[] = '</label> &nbsp; ';
                    }
                    $aFields[] = substr(implode('', $a), 0, -8) . '</div>';
                    break;
                case 'color':
                    $aFields[] = '<input type="' . $aField['type'] . '" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aField['value'] . '"' . (($aField['min'] !== null) ? ' min="' . $aField['min'] . '"' : '') . (($aField['max'] !== null) ? ' max="' . $aField['max'] . '"' : '') . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>';
                    break;
                case 'date':
                    $aFields[] = '<input type="' . $aField['type'] . '" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aField['value'] . '"' . (($aField['min'] !== null) ? ' min="' . $aField['min'] . '"' : '') . (($aField['max'] !== null) ? ' max="' . $aField['max'] . '"' : '') . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>';
                    break;
                case 'time':
                    $aFields[] = '<input type="' . $aField['type'] . '" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aField['value'] . '"' . (($aField['min'] !== null) ? ' min="' . $aField['min'] . '"' : '') . (($aField['max'] !== null) ? ' max="' . $aField['max'] . '"' : '') . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>';
                    break;
                case 'range':
                    $a = array();
                    $a[] = '<label id="' . $iID . '-label"' . ((strlen($aField['descriptiveLabelClass'])) ? ' class="' . $aField['descriptiveLabelClass'] . '"' : '') . '>';
                    if (($aField['labelPosition'] == 'after') && (strlen($aField['descriptiveLabel']))) {
                        $a[] = '<span class="description">' . $aField['descriptiveLabel'] . ' </span>';
                    }
                    $a[] = '<input type="' . $aField['type'] . '" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ' style="vertical-align: text-bottom;' . ((strlen($aField['additionalCss'])) ? ' ' . $aField['additionalCss'] : '') . '" value="' . $aField['value'] . '"' . (($aField['min'] !== null) ? ' min="' . $aField['min'] . '"' : '') . (($aField['max'] !== null) ? ' max="' . $aField['max'] . '"' : '') . (($aField['step']) ? ' step="' . $aField['step'] . '"' : '') . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>';
                    if (($aField['labelPosition'] == 'before') && (strlen($aField['descriptiveLabel']))) {
                        $a[] = '<span class="description"> ' . $aField['descriptiveLabel'] . '</span>';
                    }
                    $a[] = '</label>';
                    $aFields[] = implode(' ', $a);
                    break;
                case 'code':
                    $aFields[] = $aField['value'] . "\n";
                    break;
                default: /* text|password|name|file */
                    $aFields[] = '<input type="' . $aField['type'] . '" id="' . $iID . '" name="' . $aField['input'] . '"' . ((strlen($aField['class'])) ? ' class="' . $aField['class'] . '"' : '') . ' data-type="' . $aField['type'] . '"' . ((count($aRestrictions)) ? ' data-restrictions="' . implode('|', $aRestrictions) . '"' : '') . ((strlen($aField['additionalCss'])) ? ' style="' . $aField['additionalCss'] . '"' : '') . ' value="' . $aField['value'] . '"' . (($aField['min'] !== null) ? ' min="' . $aField['min'] . '"' : '') . (($aField['max'] !== null) ? ' max="' . $aField['max'] . '"' : '') . (($aField['filters']) ? ' accept="' . implode(',', $aField['filters']) . '"' : '') . (($aField['required']) ? ' required' : '') . (($aField['placeholder']) ? ' placeholder="' . $aField['placeholder'] . '"' : '') . '>';
                    break;
            }
            if ($aField['type'] == 'checkboxGroup') {
                $iID = '';
            }
            else {
                $iID = 'F-' . str_replace(array('[', ']'), '', $aField['input']) . '-' . $i;
            }
            if (($aField['labelPosition'] == 'after') && (strlen($aField['label']))) {
                $aFields[] = '<label' . ((strlen($iID)) ? ' for="' . $iID . '"' : '') . '>' . $aField['label'] . '</label>';
            }
            if ($aField['type'] != 'code') {
                $aFields[] = '</div>' . "\n";
                $aScripts[str_replace('Group', '', $aField['type'])] = 1;
                if ($aField['type'] != 'file') {
                    $aData = array();
                    $aData['type'] = $aField['type'];
                    if ($aField['required']) {
                        $aData['required'] = '1';
                    }
                    if (($aField['type'] != 'checkbox') && ($aField['type'] != 'radio')) {
                        if ($aField['min'] !== null) {
                            $aData['min'] = $aField['min'];
                        }
                        if ($aField['max'] !== null) {
                            $aData['max'] = $aField['max'];
                        }
                        if (count($aRestrictions)) {
                            $aData['restrictions'] = implode('¦', $aRestrictions);
                        }
                        if ($aField['countrycode'] !== null) {
                            $aData['countrycode'] = $aField['countrycode'];
                        }
                    }
                    $aHiddens[] = '<input type="hidden" name="' . $aField['input'] . '-data" value=\'' . str_replace("'", "\'", json_encode($aData, JSON_UNESCAPED_UNICODE)) . '\'>';
                }
            }
        }

        # Buttons
        if (count($this->aButtons)) {
            $aFields[] = '<div class="row buttons">' . "\n";
            foreach ($this->aButtons as $aButton) {
                if ((strlen($aButton['label'])) && ($aButton['labelPosition'] == 'before')) {
                    $aFields[] = '<label' . ((strlen($aButton['labelClass'])) ? ' class="' . $aButton['labelClass'] . '"' : '') . '>' . trim($aButton['label']) . '</label>' . "\n";
                }
                $aFields[] = '<input type="' . $aButton['type'] . '"' . ((strlen($aButton['id'])) ? ' id="' . $aButton['id'] . '"' : '') . ((strlen($aButton['class'])) ? ' class="' . $aButton['class'] . '"' : '') . ((strlen($aButton['name'])) ? ' name="' . $aButton['name'] . '"' : '') . ((strlen($aButton['value'])) ? ' value="' . $aButton['value'] . '"' : '') . ((strlen($aButton['additionalCss'])) ? ' style="' . $aButton['additionalCss'] . '"' : '') . '>' . "\n";
                if ((strlen($aButton['label'])) && ($aButton['labelPosition'] == 'after')) {
                    $aFields[] = '<label' . ((strlen($aButton['labelClass'])) ? ' class="' . $aButton['labelClass'] . '"' : '') . '>' . trim($aButton['label']) . '</label>' . "\n";
                }
            }
            $aFields[] = '</div>' . "\n";
        }

        # Output
        echo "\n" .
            '<div class="content">' . "\n" .
                '<form id="' . ((strlen($this->aForm['id'])) ? $this->aForm['id'] : 'Form-0') . '"' . ((strlen($this->aForm['class'])) ? ' class="' . $this->aForm['class'] . '"' : '') . ((strlen($this->aForm['name'])) ? ' name="' . $this->aForm['name'] . '"' : '') . ' method="' . $this->aForm['method'] . '" enctype="multipart/form-data" action="' . $this->aForm['url'] . '">' . "\n" .
                    '<div class="hiddens">' . "\n" .
                        implode("\n", $aHiddens) . "\n" .
                    '</div>' . "\n" .
                    implode('', $aFields) . "\n" .
                '</form>' . "\n" .
                '<script>';
        foreach ($aScripts as $key => $value) {
            echo "\n    " . $this->getScript($key, $value);
        }
        echo "\n" . 
                '</script>' . "\n" .
            '</div>' . "\n";
    }

    /**
     * protected function getScript()
     * 
     * Delivers the script for checking the fields of a specific type
     * 
     * @param string $type The type of field to deliver the script for
     * @return string The requested script
     */
    protected function getScript(string $type, string $param = ''):string {
        $script = '';
        switch ($type) {
            case 'preload':
                $script = '$(document).ready(function() {
                    $("#' . $param . ' *").each(function() {
                        if ($(this).attr("required")) {
                            $(this).removeAttr("required").addClass("required");
                        }
                    });
                });
                jQuery.prototype.hasAttr = function(attribute) {
                    var a = $(this).attr(attribute);
                    return ((typeof a !== "undefined") && (a !== false));
                };
                function str_word_count(string) {
                    var t = new Array(" ", "\t", "=", "+", "-", "*", "/", "\\\\", ",", ".", ";", ":", "[", "]", "{", "}", "(", ")", "<", ">", "&", "%", "$", "@", "#", "^", "!", "?", "~");
                    string = string.replace(t.join("|"), " ").trim();
                    num = 0;
                    if (string.length) {
                        word_array = string.split(" ");
                        num = word_array.length;
                    }
                    return num;
                }
                function executeOperation(operator, value1, value2, funct = "") {
                    if (funct.length) {
                        switch (funct) {
                            case "strlen":
                                value1 = value1.length;
                                break;
                            case "str_word_count":
                                value1 = str_word_count(value1);
                                break;
                        }
                    }
                    switch (operator) {
                        case "==": return (value1 == value2);
                        case "===": return (value1 === value2);
                        case "!=": return (value1 != value2);
                        case "!==": return (value1 !== value2);
                        case ">": return (value1 > value2);
                        case ">=": return (value1 >= value2);
                        case "<": return (value1 < value2);
                        case "<=": return (value1 <= value2);
                    }
                }
                function validateIp(ip) {
                    var ipv4 = /^((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)(\.(?!$)|$)){4}$/i;
                    var ipv6 = /(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))/i;
                    if (ipv4.test(ip)) {
                        return true;
                    } else if (ipv6.test(ip)) {
                        return true;
                    }
                    return false;
                }';
                break;
            case 'form':
                $script = '$("#' . $param . '").submit(function() {
                    var sbmt = $(this).find("input[type=submit]");
                    sbmt.prop("disabled", true);
                    $("*").css("cursor", "wait");
                    var ret = true;
                    /* Separate iterations because of radio groups */
                    $("input,textarea,select", "#' . $param . '").not("[type=\'hidden\'],[type=\'button\'],[type=\'submit\'],[type=\'reset\']").each(function() {
                        if (!(($(this).hasClass("correct")) || ($(this).hasClass("wrong")))) {
                            $(this).trigger("blur");
                        }
                    });
                    $("input,textarea,select,.radiogroup", "#' . $param . '").not("[type=\'hidden\'],[type=\'button\'],[type=\'submit\'],[type=\'reset\']").each(function() {
                        if ($(this).hasClass("wrong")) {
                            ret = false;
                        }
                    });
                    if (!ret) {
                        sbmt.prop("disabled", false);
                        $("*").css("cursor", "");
                    } else {
                        $(this).append(\'<input type="hidden" name="\' + sbmt.attr("name") + \'" value="\' + sbmt.val() + \'">\');
                    }
                    return ret;
                });';
                break;
            case 'bic': /* incl restrictions */
                $script = '$("input[data-type=\'bic\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        value = $(this).val().toUpperCase().replace(/ /g, "");
                        var rx = /^[A-Z]{6}[0-9A-Z]{2}(?:[0-9A-Z]{3})?$/;
                        var ret = rx.test(value);
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'checkbox': /* required only */
                $script = '$("input[data-type=\'checkbox\']").on("blur click", function() {
                    if ($(this).hasClass("required")) {
                        if ($(this).prop("checked")) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'creditcard': /* incl restrictions */
                $script = '$("input[data-type=\'creditcard\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        var value = $(this).val().replace(/ /g, "");
                        var rx = /^[3-6]{1}[0-9]{12,18}$/;
                        if (rx.test(value)) {
                            var digits = [];
                            var j = 1, digit = "";
                            for (var i = value.length - 1; i >= 0; i--) {
                                if ((j % 2) == 0) {
                                    digit = (parseInt(value.charAt(i), 10) * 2);
                                    digits[digits.length] = digit.toString().charAt(0);
                                    if (digit.toString().length == 2) {
                                        digits[digits.length] = digit.toString().charAt(1);
                                    }
                                } else {
                                    digit = value.charAt(i);
                                    digits[digits.length] = digit;
                                }
                                j++;
                            }
                            var sum = 0;
                            for (i = 0; i < digits.length; i ++) {
                                sum += parseInt(digits[i], 10);
                            }
                            if ((sum % 10) == 0) {
                                ret = true;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'color': /* required only, incl restrictions */
                $script = '$("input[data-type=\'color\']").on("blur keyup",function() {
                    ret = true;
                    if ($(this).hasClass("required")) {
                        if (($(this).val() == "#000000") || ($(this).val() == "#ffffff")) {
                            ret = false;
                        }
                    }
                    if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                        var restrictions = $(this).attr("data-restrictions");
                        restrictions = restrictions.split("|");
                        for (var i = 0; i < restrictions.length; i ++) {
                            var restriction = restrictions[i].split("~");
                            if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                ret = false;
                            }
                        }
                    }
                    if (ret) {
                        $(this). addClass("correct").removeClass("wrong");
                    } else {
                        $(this).addClass("wrong").removeClass("correct");
                    }
                });';
                break;
            case 'currency': /* incl min|max|restrictions */
                $script = '$("input[data-type=\'currency\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        var rx = /^(\-|\+)?\d+(?:\.\d{2})?$/;
                        if (rx.test($(this).val())) {
                            var ret = parseFloat($(this).val());
                            if ((!isNaN(ret)) && (ret == $(this).val())) {
                                ret = true;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val() < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val() > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this). addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'date': /* incl min|max|restrictions */
                $script = '$("input[data-type=\'date\']").on("blur keyup change",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = ($(this).val().length > 0);
                        var date1 = new Date($(this).val());
                        if ($(this).hasAttr("min")) {
                            var date2 = new Date($(this).attr("min"));
                            if (date1 < date2) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            var date2 = new Date($(this).attr("max"));
                            if (date1 > date2) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'email': /* incl min|max|restrictions */
                $script = '$("input[data-type=\'email\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        var rx = /^[\w\.-]+@[\w\.-]+\.\w+$/i;
                        if (rx.test($(this).val())) {
                            ret = true;
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val().length < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val().length > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this). addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'file': /* required and accept */
                $script = '$("input[data-type=\'file\']").on("blur change",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        ret = false;
                        if ($(this).val().length > 0) {
                            ret = true;
                        }
                        if ((ret) && ($(this).hasAttr("accept"))) {
                            var rx = new RegExp(".+(" + $(this).attr("accept").replace(/,/g, "|") + ")$");
                            if (!$(this).val().match(rx)) {
                                ret = false;
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'float': /* incl min|max|restrictions */
                $script = '$("input[data-type=\'float\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        var rx = /^[\+\-]?\d+(?:\.\d+)?([eE]{0,1}[\-\+]?\d+)?$/;
                        if (rx.test($(this).val())) {
                            var r = parseFloat($(this).val());
                            if ((!isNaN(r)) && (r == $(this).val())) {
                                ret = true;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val() < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val() > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this). addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'iban': /* incl restrictions */
                $script = '$("input[data-type=\'iban\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        ret = false;
                        if ($(this).val().length > 0) {
                            value = $(this).val().toUpperCase().replace(/ /g, "");
                            var BBAN = value.substr(4);
                            var CTY = (value.charCodeAt(0) - 55) + "" + (value.charCodeAt(1) - 55);
                            var CHK = value.substr(2, 2);
                            var check = (BBAN + CTY + CHK);
                            var len = 0;
                            var run = "";
                            while (check.length > 0) {
                                len = (9 - run.toString().length);
                                run = (Number(run + check.substr(0, len)) % 97);
                                check = check.substr(len);
                            }
                            ret = (run == 1);
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).val($(this).val().toUpperCase()).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'integer': /* incl min|max|restrictions */
                $script = '$("input[data-type=\'integer\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        var rx = /^[\-\+]?\d*$/;
                        if (rx.test($(this).val())) {
                            var r = parseInt($(this).val(), 10);
                            if ((!isNaN(r)) && (r == $(this).val())) {
                                ret = true;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val() < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val() > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this). addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'ip': /* v4|v6|min|max|restrictions */
                $script = '$("input[data-type=\'ip\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        if ($(this).val().length > 0) {
                            ret = validateIp($(this).val());
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val().length < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val().length > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'name': /* required only, incl min|max|restrictions */
                $script = '$("input[data-type=\'name\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        if ($(this).val().length > 0) {
                            ret = true;
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val().length < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val().length > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'password':
                $script = '$("input[data-type=\'password\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        if ($(this).val().length > 0) {
                            ret = true;
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val().length < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val().length > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'radio': /* required only */
                $script = '$("input[data-type=\'radio\']").on("blur click", function() {
                    var name = $(this).attr("name");
                    if ($(this).hasClass("required")) {
                        var ret = false;
                        $("input[name=\'" + name + "\']").each(function() {
                            if ($(this).prop("checked")) {
                                ret = true;
                            }
                        });
                        if (ret) {
                            if ($(this).parent().parent().hasClass("radiogroup")) {
                                $(this).parent().parent().addClass("correct").removeClass("wrong");
                            }
                            else {
                                $("input[name=\'" + name + "\']").removeClass("correct").removeClass("wrong");
                                $("input[name=\'" + name + "\']:checked").addClass("correct");
                            }
                        } else {
                            if ($(this).parent().parent().hasClass("radiogroup")) {
                                $(this).parent().parent().addClass("wrong").removeClass("correct");
                            }
                            else {
                                $("input[name=\'" + name + "\']").addClass("wrong").removeClass("correct");
                            }
                        }
                    } else {
                        if ($(this).parent().parent().hasClass("radiogroup")) {
                            $(this).parent().parent().addClass("correct").removeClass("wrong");
                        }
                        else {
                            $("input[name=\'" + name + "\']").addClass("correct").removeClass("wrong");
                        }
                    }
                });';
                break;
            case 'range': /* required only, incl min|max|restrictions */
                $script = '$("input[data-type=\'range\']").on("blur input", function() {
                    var id = $(this).attr("id");
                    $("#" + id + "-label span").text($(this).val());
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = true;
                        if ($(this).hasClass("required")) {
                            ret = ($(this).val() != 0);
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val() < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val() > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'select': /* required only, incl restrictions */
                $script = '$("select[data-type=\'select\']").on("blur change", function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        if ($(this).val().length > 0) {
                            ret = true;
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'tel': /* incl min|max|restrictions */
                $script = '$("input[data-type=\'tel\']").on("blur keyup",function() {
                    var rx = /^(\+[0-9]{2,3}[\d\s\/\(\)\ ].+[0-9]{2,5}|0[0-9]{2,5})[\d\s\/\(\)\-\ ].+[0-9]{2,10}$/;
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        if (rx.test($(this).val())) {
                            ret = true;
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val().length < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val().length > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this). addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'text': /* required only, incl min|max|restrictions */
                $script = '$("input[data-type=\'text\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        if ($(this).val().length > 0) {
                            ret = true;
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val().length < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val().length > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'textarea': /* required only, incl min|max|restrictions */
                $script = '$("textarea[data-type=\'textarea\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        if ($(this).val().length > 0) {
                            ret = true;
                        }
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val().length < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val().length > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'time': /* incl min|max|restrictions */
                $script = '$("input[data-type=\'time\']").on("blur keyup change",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = true;
                        var time = $(this).val().split(":");
                        if ((time.length < 2) || (time.length > 3)) {
                            ret = false;
                        }
                        else if ((Number(time[0]) < 0) || (Number(time[0]) > 24)) {
                            ret = false;
                        }
                        else if ((Number(time[1]) < 0) || (Number(time[1]) > 59)) {
                            ret = false;
                        }
                        else if ((time.length == 3) && ((Number(time[2]) < 0) || (Number(time[2]) > 59))) {
                            ret = false;
                        }
                        var time1 = Number(time.join(""));
                        if ($(this).hasAttr("min")) {
                            var time2 = (Number($(this).attr("min").split(":").join("")) / 100);
                            if (time1 < time2) {
                                ret = false;
                            }
                        }
                        if ($(this).hasAttr("max")) {
                            var time2 = (Number($(this).attr("max").split(":").join("")) / 100);
                            if (time1 > time2) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'url': /* incl min|max|restrictions */
                $script = '$("input[data-type=\'url\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var rx = /^(?:(?:https?|ftp):\/\/|www\.)[-a-z0-9+&@#%?=~_|!:,;]{2,}(\.[-a-z0-9+&@#\/%=~_|]{2,}){2,}$/i;
                        var ret = $(this).val().match(rx);
                        if ((ret) && ($(this).hasAttr("min"))) {
                            if ($(this).val().length < Number($(this).attr("min"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("max"))) {
                            if ($(this).val().length > Number($(this).attr("max"))) {
                                ret = false;
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
            case 'zipcode': /* uses attrib "data-countrycode", else required only. incl restrictions */
                $script = '$("input[data-type=\'zipcode\']").on("blur keyup",function() {
                    if (($(this).hasClass("required")) || ($(this).val().length)) {
                        var ret = false;
                        if ($(this).val().length > 0) {
                            ret = true;
                        }
                        if ((ret) && ($(this).hasAttr("data-countrycode")) && ($(this).attr("data-countrycode").length)) {
                            var rx = "";
                            switch ($(this).attr("data-countrycode").toUpperCase()) {
                                case "AD":
                                    rx = /^AD\d{3}$/;
                                    break;
                                case "AE":
                                    /* No postal codes */
                                    break;
                                case "AF":
                                    rx = /^([1-2]\d|3[0-4])\d{2}$/;
                                    break;
                                case "AG":
                                    /* No postal codes */
                                    break;
                                case "AI":
                                    return ($zipCode == /AI-2640/); /* single code for all addresses */
                                    break;
                                case "AL":
                                    rx = /^\d{4}$/;
                                    break;
                                case "AM":
                                    rx = /^\d{4}(?:\d{2})?$/;
                                    break;
                                case "AO":
                                    /* No postal codes */
                                    break;
                                case "AQ":
                                    return ($zipCode == /BIQQ 1ZZ/); /* single code for all addresses */
                                    break;
                                case "AR":
                                    rx = /^(\d{4}|[A-H]|[J-N]|[P-Z]\d{4}[A-Z]{3})$/;
                                    break;
                                case "AS":
                                    rx = /^\d{5}(?:\-\d{4})?$/;
                                    break;
                                case "AT":
                                    rx = /^[1-9]\d{3}$/;
                                    break;
                                case "AU":
                                    rx = /^\d{4}$/;
                                    break;
                                case "AW":
                                    /* No postal codes */
                                    break;
                                case "AX":
                                    rx = /^(?:AX\-)?22\d{3}$/;
                                    break;
                                case "AZ":
                                    rx = /^(AZ\s\d{4}|\d{6})$/;
                                    break;
                                case "BA":
                                    rx = /^\d{5}$/;
                                    break;
                                case "BB":
                                    rx = /^BB\d{5}$/;
                                    break;
                                case "BD":
                                    rx = /^\d{4}$/;
                                    break;
                                case "BE":
                                    rx = /^\d{4}$/;
                                    break;
                                case "BF":
                                    /* No postal codes */
                                    break;
                                case "BG":
                                    rx = /^\d{4}$/;
                                    break;
                                case "BH":
                                    rx = /^([1-9](0[1-9]|1\d)|1([0-1]\d{2}|2[0-1][0-6]))$/;
                                    break;
                                case "BI":
                                    /* No postal codes */
                                    break;
                                case "BJ":
                                    /* No postal codes */
                                    break;
                                case "BL":
                                    return ($zipCode == "97133"); /* single code for all addresses */
                                    break;
                                case "BM":
                                    rx = /^[A-Z]{2}\s([BG]X|\d{2})$/;
                                    break;
                                case "BN":
                                    rx = /^[A-Z]{2}\d{4}$/;
                                    break;
                                case "BO":
                                    /* No postal codes */
                                    break;
                                case "BQ":
                                    /* No postal codes */
                                    break;
                                case "BR":
                                    rx = /^\d{5}(?:\-\d{3})?$/;
                                    break;
                                case "BS":
                                    /* No postal codes */
                                    break;
                                case "BT":
                                    rx = /^\d{5}$/;
                                    break;
                                case "BW":
                                    /* No postal codes */
                                    break;
                                case "BY":
                                    rx = /^\d{6}$/;
                                    break;
                                case "BZ":
                                    /* No postal codes */
                                    break;
                                case "CA":
                                    rx = /^([A-C]|E|[G-H]|[J-N]|P|[R-Z])\d([A-C]|E|[G-H]|[J-N]|P|[R-Z])\s\d([A-C]|E|[G-H]|[J-N]|P|[R-Z])\d$/;
                                    break;
                                case "CC":
                                    rx = /^\d{4}$/;
                                    break;
                                case "CD":
                                    /* No postal codes */
                                    break;
                                case "CF":
                                    /* No postal codes */
                                    break;
                                case "CG":
                                    /* No postal codes */
                                    break;
                                case "CH":
                                    rx = /^\d{4}$/;
                                    break;
                                case "CI":
                                    /* No postal codes */
                                    break;
                                case "CK":
                                    /* No postal codes */
                                    break;
                                case "CL":
                                    rx = /^\d{3}(?:\-)?\d{4}$/;
                                    break;
                                case "CM":
                                    /* No postal codes */
                                    break;
                                case "CN":
                                    rx = /^\d{6}$/;
                                    break;
                                case "CO":
                                    rx = /^([0-2]\d|3[0-2])\d{4}$/;
                                    break;
                                case "CR":
                                    rx = /^(\d{4}|\d{5}(?:\-\d{4})?)$/;
                                    break;
                                case "CU":
                                    rx = /^\d{5}$/;
                                    break;
                                case "CV":
                                    rx = /^\d{4}$/;
                                    break;
                                case "CW":
                                    /* No postal codes */
                                    break;
                                case "CX":
                                    rx = /^\d{4}(?:\d)?$/;
                                    break;
                                case "CZ":
                                    rx = /^[1-7]\d{2}\s\d{2}$/;
                                    break;
                                case "DE":
                                    rx = /^\d{5}$/g;
                                    break;
                                case "DJ":
                                    /* No postal codes */
                                    break;
                                case "DK":
                                    rx = /^\d{5}$/;
                                    break;
                                case "DM":
                                    /* No postal codes */
                                    break;
                                case "DZ":
                                    rx = /^([0-3]\d|4[0-8])\d{3}$/;
                                    break;
                                case "EC":
                                    rx = /^\d{6}$/;
                                    break;
                                case "EE":
                                    rx = /^\d{5}$/;
                                    break;
                                case "ER":
                                    /* No postal codes */
                                    break;
                                case "ES":
                                    rx = /^([0-4]\d|5[0-2])\d{3}$/;
                                    break;
                                case "ET":
                                    rx = /^\d{4}$/;
                                    break;
                                case "FI":
                                    rx = /^\d{4}[0-1]$/;
                                    break;
                                case "FJ":
                                    /* No postal codes */
                                    break;
                                case "FK":
                                    return ($zipCode == "FIQQ 1ZZ"); /* single code for all addresses */
                                    break;
                                case "FM":
                                    rx = /^\d{5}(?:\-\d{4})?$/;
                                    break;
                                case "FO":
                                    rx = /^FO\-\d{3}$/;
                                    break;
                                case "FR":
                                    rx = /^\d{5}$/;
                                    break;
                                case "GA":
                                    /* No postal codes */
                                    break;
                                case "GB":
                                    rx = /^[A-Z]{1,2}\d[\dA-Z]?\s?\d[A-Z]{2}$/;
                                    break;
                                case "GD":
                                    /* No postal codes */
                                    break;
                                case "GE":
                                    rx = /^\d{4}$/;
                                    break;
                                case "GF":
                                    rx = /^973([0-8]\d|90)$/;
                                    break;
                                case "GG":
                                    rx = /^GY\d{1,2}\s\d[A-Z]{2}$/;
                                    break;
                                case "GH":
                                    rx = /^[A-Z]{2}\-\d{4}(?:\-\d{4})?$/;
                                    break;
                                case "GI":
                                    return ($zipCode == "GX11 1AA"); /* single code for all addresses */
                                    break;
                                case "GL":
                                    rx = /^\d{4}$/;
                                    break;
                                case "GM":
                                    /* No postal codes */
                                    break;
                                case "GN":
                                    rx = /^\d{3}$/;
                                    break;
                                case "GP":
                                    rx = /^971([0-8]\d|90)$/;
                                    break;
                                case "GQ":
                                    /* No postal codes */
                                    break;
                                case "GR":
                                    rx = /^\d{3}\s\d{2}$/;
                                    break;
                                case "GS":
                                    return ($zipCode == "SIQQ 1ZZ"); /* single code for all addresses */
                                    break;
                                case "GT":
                                    rx = /^\d{5}$/;
                                    break;
                                case "GU":
                                    rx = /^969([1-2]\d|3[1-2])$/;
                                    break;
                                case "GW":
                                    rx = /^\d{4}$/;
                                    break;
                                case "GY":
                                    /* No postal codes */
                                    break;
                                case "HK":
                                    /* No postal codes */
                                    break;
                                case "HM":
                                    /* No postal codes */
                                    break;
                                case "HN":
                                    rx = /^([A-Z]{2}\d{4}|\d{5})$/;
                                    break;
                                case "HR":
                                    rx = /^\d{5}$/;
                                    break;
                                case "HT":
                                    rx = /^\d{4}$/;
                                    break;
                                case "HU":
                                    rx = /^\d{4}$/;
                                    break;
                                case "ID":
                                    rx = /^\d{5}$/;
                                    break;
                                case "IE":
                                    rx = /^(A|[C-F]|H|K|N|P|R|T|[V-X])\d{2}\s\w{4}$/;
                                    break;
                                case "IL":
                                    rx = /^\d{5}(?:\d{2})?$/;
                                    break;
                                case "IM":
                                    rx = /^IM\d(?:\d)?\s\d[A-Z]{2}$/;
                                    break;
                                case "IN":
                                    rx = /^\d{3}(?:\s)?\d{3}$/;
                                    break;
                                case "IO":
                                    return ($zipCode == "BBND 1ZZ"); /* single code for all addresses */
                                    break;
                                case "IQ":
                                    rx = /^\d{5}$/;
                                    break;
                                case "IR":
                                    rx = /^\d{5}\-\d{5}$/;
                                    break;
                                case "IS":
                                    rx = /^\d{3}$/;
                                    break;
                                case "IT":
                                    rx = /^\d{5}$/;
                                    break;
                                case "JE":
                                    rx = /^JE\d(?:\d)?\s\d[A-Z]{2}$/;
                                    break;
                                case "JM":
                                    rx = /^\d{1,2}$/;
                                    break;
                                case "JO":
                                    rx = /^\d{5}$/;
                                    break;
                                case "JP":
                                    rx = /^\d{3}\-\d{4}$/;
                                    break;
                                case "KE":
                                    rx = /^\d{5}$/;
                                    break;
                                case "KG":
                                    rx = /^\d{6}$/;
                                    break;
                                case "KH":
                                    return ($zipCode == "120000"); /* single code for all addresses */
                                    break;
                                case "KI":
                                    /* No postal codes */
                                    break;
                                case "KM":
                                    /* No postal codes */
                                    break;
                                case "KN":
                                    rx = /^\KN\d{4}$/;
                                    break;
                                case "KP":
                                    /* No postal codes */
                                    break;
                                case "KR":
                                    rx = /^\d{5}$/;
                                    break;
                                case "KW":
                                    rx = /^\d{5}$/;
                                    break;
                                case "KY":
                                    rx = /^\KY\d\-\d{4}$/;
                                    break;
                                case "KZ":
                                    rx = /^\d{6}$/;
                                    break;
                                case "LA":
                                    rx = /^\d{5}$/;
                                    break;
                                case "LB":
                                    rx = /^(\d{5}|\d{4}\s\d{4})$/;
                                    break;
                                case "LC":
                                    rx = /^LC\d{2}\s\d{3}$/;
                                    break;
                                case "LI":
                                    rx = /^94(8[5-9]|9[0-8])$/;
                                    break;
                                case "LK":
                                    rx = /^\d{5}$/;
                                    break;
                                case "LR":
                                    rx = /^\d{2}(?:\d{2})?$/;
                                    break;
                                case "LS":
                                    rx = /^\d{3}$/;
                                    break;
                                case "LT":
                                    rx = /^LT\-\d{5}$/;
                                    break;
                                case "LU":
                                    rx = /^\d{4}$/;
                                    break;
                                case "LV":
                                    rx = /^LV\-\d{4}$/;
                                    break;
                                case "LY":
                                    /* No postal codes */
                                    break;
                                case "MA":
                                    rx = /^\d{5}$/;
                                    break;
                                case "MC":
                                    rx = /^\980d{2}$/;
                                    break;
                                case "MD":
                                    rx = /^\MD(?:\-)?d{4}$/;
                                    break;
                                case "ME":
                                    rx = /^\d{4}$/;
                                    break;
                                case "MF":
                                    return ($zipCode == "97150"); /* single code for all addresses */
                                    break;
                                case "MG":
                                    rx = /^\d{3}$/;
                                    break;
                                case "MH":
                                    rx = /^\969(6\d|70)(?:\-\d{4})?$/;
                                    break;
                                case "MK":
                                    rx = /^\d{4}$/;
                                    break;
                                case "ML":
                                    /* No postal codes */
                                    break;
                                case "MM":
                                    rx = /^\d{5}$/;
                                    break;
                                case "MN":
                                    rx = /^\d{5}$/;
                                    break;
                                case "MO":
                                    /* No postal codes */
                                    break;
                                case "MP":
                                    rx = /^\9695[0-2](?:\-\d{4})?$/;
                                    break;
                                case "MQ":
                                    rx = /^\972([0-8]\d|90)$/;
                                    break;
                                case "MR":
                                    /* No postal codes */
                                    break;
                                case "MS":
                                    rx = /^MSR\s1(1[1-9]\d|2\d{2}|3[0-4]\d|350)$/;
                                    break;
                                case "MT":
                                    rx = /^[A-Z]{3}\s\d{4}$/;
                                    break;
                                case "MU":
                                    rx = /^\d{5}$/;
                                    break;
                                case "MV":
                                    rx = /^\d{5}$/;
                                    break;
                                case "MW":
                                    rx = /^\d{6}$/;
                                    break;
                                case "MX":
                                    rx = /^\d{5}$/;
                                    break;
                                case "MY":
                                    rx = /^\d{5}$/;
                                    break;
                                case "MZ":
                                    rx = /^\d{4}$/;
                                    break;
                                case "NA":
                                    rx = /^\d{5}$/;
                                    break;
                                case "NC":
                                    rx = /^988([0-8]\d|90)$/;
                                    break;
                                case "NE":
                                    rx = /^\d{4}$/;
                                    break;
                                case "NF":
                                    rx = /^\d{4}$/;
                                    break;
                                case "NG":
                                    rx = /^\d{5}$/;
                                    break;
                                case "NI":
                                    rx = /^\d{4}$/;
                                    break;
                                case "NL":
                                    rx = /^\d{4}(?:\s{1})?[A-Z]{2}$/;
                                    break;
                                case "NO":
                                    rx = /^\d{4}$/;
                                    break;
                                case "NP":
                                    rx = /^\d{5}$/;
                                    break;
                                case "NR":
                                    /* No postal codes */
                                    break;
                                case "NU":
                                    /* No postal codes */
                                    break;
                                case "NZ":
                                    rx = /^\d{4}$/;
                                    break;
                                case "OM":
                                    rx = /^\d{3}$/;
                                    break;
                                case "PA":
                                    rx = /^\d{4}$/;
                                    break;
                                case "PE":
                                    rx = /^(\d{5}|PE\s\d{4})$/;
                                    break;
                                case "PF":
                                    rx = /^987([0-8]\d|90)$/;
                                    break;
                                case "PG":
                                    rx = /^\d{3}$/;
                                    break;
                                case "PH":
                                    rx = /^\d{4}$/;
                                    break;
                                case "PK":
                                    rx = /^\d{5}$/;
                                    break;
                                case "PG":
                                    rx = /^\d{2}\-\d{3}$/;
                                    break;
                                case "PM":
                                    return ($zipCode == "97500"); /* single code for all addresses */
                                    break;
                                case "PN":
                                    return ($zipCode == "PCRN 1ZZ"); /* single code for all addresses */
                                    break;
                                case "PR":
                                    rx = /^00[6|7|9]\d{2}(?:\-\d{4})?$/;
                                    break;
                                case "PS":
                                    rx = /^\d{3}$/;
                                    break;
                                case "PT":
                                    rx = /^\d{4}(?:\-\d{3})?$/;
                                    break;
                                case "PW":
                                    rx = /^969[39|40](?:\-\d{4})?$/;
                                    break;
                                case "PY":
                                    rx = /^\d{4}$/;
                                    break;
                                case "QA":
                                    /* No postal codes */
                                    break;
                                case "RE":
                                    rx = /^974([0-8]\d|90)$/;
                                    break;
                                case "RO":
                                    rx = /^\d{6}$/;
                                    break;
                                case "RS":
                                    rx = /^\d{5}$/;
                                    break;
                                case "RU":
                                    rx = /^\d{6}$/;
                                    break;
                                case "RW":
                                    /* No postal codes */
                                    break;
                                case "SA":
                                    rx = /^\d{5}(?:\-\d{4})?$/;
                                    break;
                                case "SB":
                                    /* No postal codes */
                                    break;
                                case "SC":
                                    /* No postal codes */
                                    break;
                                case "SD":
                                    rx = /^\d{5}$/;
                                    break;
                                case "SE":
                                    rx = /^\d{3}\s\d{2}$/;
                                    break;
                                case "SG":
                                    rx = /^\d{2}(?:\d{2})?(?:\d{2})?$/;
                                    break;
                                case "SH":
                                    rx = /^[STHL|ASCN|TDCU]\s1ZZ$/;
                                    break;
                                case "SI":
                                    rx = /^(?:SI\s)?\d{4}$/;
                                    break;
                                case "SJ":
                                    rx = /^\d{4}$/;
                                    break;
                                case "SK":
                                    rx = /^\d{3}\s\d{2}$/;
                                    break;
                                case "SL":
                                    /* No postal codes */
                                    break;
                                case "SM":
                                    rx = /^4789\d$/;
                                    break;
                                case "SN":
                                    rx = /^\d{5}$/;
                                    break;
                                case "SO":
                                    rx = /^[A-Z]{2}\s\d{5}$/;
                                    break;
                                case "SR":
                                    /* No postal codes */
                                    break;
                                case "SS":
                                    /* No postal codes */
                                    break;
                                case "ST":
                                    /* No postal codes */
                                    break;
                                case "SV":
                                    rx = /^\d{4}$/;
                                    break;
                                case "SX":
                                    /* No postal codes */
                                    break;
                                case "SY":
                                    rx = /^\d{4}$/;
                                    break;
                                case "SZ":
                                    rx = /^[H|M|S|L]\d{3}$/;
                                    break;
                                case "TC":
                                    return ($zipCode == "TKCA 1ZZ"); /* single code for all addresses */
                                    break;
                                case "TD":
                                    /* No postal codes */
                                    break;
                                case "TF":
                                    /* No postal codes */
                                    break;
                                case "TG":
                                    /* No postal codes */
                                    break;
                                case "TH":
                                    rx = /^\d{5}$/;
                                    break;
                                case "TJ":
                                    rx = /^\d{6}$/;
                                    break;
                                case "TK":
                                    /* No postal codes */
                                    break;
                                case "TL":
                                    /* No postal codes */
                                    break;
                                case "TM":
                                    rx = /^\d{6}$/;
                                    break;
                                case "TN":
                                    rx = /^\d{4}$/;
                                    break;
                                case "TO":
                                    /* No postal codes */
                                    break;
                                case "TR":
                                    rx = /^\d{5}$/;
                                    break;
                                case "TT":
                                    rx = /^\d{6}$/;
                                    break;
                                case "TV":
                                    /* No postal codes */
                                    break;
                                case "TW":
                                    rx = /^\d{3}(?:\-\d{2,3})?$/;
                                    break;
                                case "TZ":
                                    rx = /^\d{5}$/;
                                    break;
                                case "UA":
                                    rx = /^\d{5}$/;
                                    break;
                                case "UG":
                                    /* No postal codes */
                                    break;
                                case "UM":
                                    return ($zipCode == "96898"); /* single code for all addresses */
                                    break;
                                case "US":
                                    rx = /^\d{5}(?:\-\d{4})?$/;
                                    break;
                                case "UY":
                                    rx = /^\d{5}$/;
                                    break;
                                case "UZ":
                                    rx = /^\d{6}$/;
                                    break;
                                case "VA":
                                    return ($zipCode == "00120"); /* single code for all addresses */
                                    break;
                                case "VC":
                                    rx = /^VC\d{4}$/;
                                    break;
                                case "VE":
                                    rx = /^\d{4}(?:\-[A-Z])?$/;
                                    break;
                                case "VG":
                                    rx = /^VG11([1-5]\d|60)$/;
                                    break;
                                case "VI":
                                    rx = /^008(0[1-9]|[1-4]\d|5[0-1])$/;
                                    break;
                                case "VN":
                                    rx = /^\d{5}$/;
                                    break;
                                case "VU":
                                    /* No postal codes */
                                    break;
                                case "WF":
                                    rx = /^986([0-8]\d|90)$/;
                                    break;
                                case "WS":
                                    rx = /^WS\d{4}$/;
                                    break;
                                case "XK":
                                    rx = /^\d{5}$/;
                                    break;
                                case "YE":
                                    /* No postal codes */
                                    break;
                                case "YT":
                                    rx = /^976([0-8]\d|90)$/;
                                    break;
                                case "ZA":
                                    rx = /^\d{5}$/;
                                    break;
                                case "ZM":
                                    rx = /^\d{5}$/;
                                    break;
                                case "ZW":
                                    /* No postal codes */
                                    break;
                            }
                            if (rx) {
                                if (!$(this).val().match(rx)) {
                                    ret = false;
                                }
                            }
                            else {
                                ret = false; /* country code gives empty regex, no postal codes exist */
                            }
                        }
                        if ((ret) && ($(this).hasAttr("data-restrictions"))) {
                            var restrictions = $(this).attr("data-restrictions");
                            restrictions = restrictions.split("|");
                            for (var i = 0; i < restrictions.length; i ++) {
                                var restriction = restrictions[i].split("~");
                                if (!executeOperation(restriction[1], $(this).val(), restriction[2], restriction[0])) {
                                    ret = false;
                                }
                            }
                        }
                        if (ret) {
                            $(this).addClass("correct").removeClass("wrong");
                        } else {
                            $(this).addClass("wrong").removeClass("correct");
                        }
                    } else {
                        $(this).addClass("correct").removeClass("wrong");
                    }
                });';
                break;
        }
        return str_replace(array("\r\n", "\r", "\n"), ' ', str_replace(array('    ', "\t"), '', $script));
    }

    /**
     * static function checkFormData()
     * 
     * Checks the entire form (all fields) for being valid
     * 
     * @param array $formData The form data array ($_POST|$_GET|$_REQUEST)
     * @return array|bool Array of fields with errors or true if all is valid
     */
    public static function checkFormData(array $formData) {
        $aControls = $aErrors = array();

        # First get the data types and restrictions
        foreach ($formData as $name => $value) {

            if (strlen($value)) {

                # Restore the array from the json data
                $value = str_replace(array('\"', "\'"), array('"', "'"), $value);
                $value = json_decode($value, true);

                if (is_array($value)) {

                    # Split into single settings
                    foreach ($value as $subkey => $subvalue) {
                        $p1 = strrpos($name, '-');
                        if ($p1 !== false) {
                            $aControls[substr($name, 0, $p1)][$subkey] = $subvalue;
                        }
                    }
                }
            }
        }
        #echo '<pre>$aControls = ' . print_r($aControls, true) . '</pre>';

        # Check the validity of each field we hold restrictions for
        $bValid = true;
        foreach ($aControls as $sControl => $aRestrictions) {
            $aRestr = array();
            if (strlen($aRestrictions['restrictions'])) {
                $r = explode('¦', $aRestrictions['restrictions']);
                foreach ($r as $restr) {
                    $aRestr[] = explode('~', $restr);
                }
            }
            if (($aRestrictions['required'] == 1) || (strlen($formData[$sControl]))) {
                $ret = html5form::checkFieldData($aRestrictions['type'], $formData[$sControl], $aRestrictions['min'], $aRestrictions['max'], $aRestr, $aRestrictions['countrycode']);
                if (!$ret) {
                    $bValid = false;
                    $aErrors[] = $sControl;
                }
            }
        }

        if ($bValid) {
            return true;
        }
        else {
            return $aErrors;
        }
    }

    /**
     * static function checkFieldData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param string $type The type of the field
     * @param $data The data entered in the field
     * @param $min The minimum value|length
     * @param $max The maximum value|length
     * @param array $restrictions Additional restrictions to be checked
     *              array(
     *                  array('function' => <Property>, 'operator' => <Operator>, 'compare' => <Compare>),
     *                  ...
     *              )
     *              'function' can be one of [value|strlen|...]
     *              'operator' can be one of [==|===|!=|!==|>|>=|<|<=]
     *              'compare' must be the value to compare against
     *              Also allowed: array(array(<Function>, <Operator>, <Compare>), ...)
     * @param string $countryCode Country code used to validate names and postal codes only
     * @return bool Is the data entered okay
     */
    public static function checkFieldData(string $type, $data, $min, $max, $restrictions = array(), $countryCode = '') {
        $ret = true;
        if ((isset($min)) && ($min !== null)) {
            switch ($type) {
                case 'currency':
                case 'integer':
                case 'float':
                    if ($data < $min) {
                        $ret = false;
                    }
                    break;
                case 'date':
                case 'time':
                    if (strtotime($data) < strtotime($min)) {
                        $ret = false;
                    }
                    break;
                default:
                    if (strlen($data) < $min) {
                        $ret = false;
                    }
                    break;
            }
        }
        if (($ret) && ((isset($max)) && ($max !== null))) {
            switch ($type) {
                case 'currency':
                case 'integer':
                case 'float':
                    if ($data > $max) {
                        $ret = false;
                    }
                    break;
                case 'date':
                case 'time':
                    if (strtotime($data) > strtotime($max)) {
                        $ret = false;
                    }
                    break;
                default:
                    if (strlen($data) > $max) {
                        $ret = false;
                    }
                    break;
            }
        }
        if ($ret) {
            foreach ($restrictions as $key => $restriction) {
                if (!array_key_exists('function', $restriction)) {
                    $restrictions[$key] = array(
                        'function' => $restriction[0],
                        'operator' => $restriction[1],
                        'compare' => $restriction[2]
                    );
                }
            }
            switch ($type) {
                case 'bic':
                    $ret = html5form::checkBicData($data, $restrictions);
                    break;
                case 'checkbox':
                    $ret = (strlen($data) > 0);
                    break;
                case 'creditcard':
                    $ret = html5form::checkCreditCardData($data, $restrictions);
                    break;
                case 'currency':
                    $ret = html5form::checkCurrencyData($data, $restrictions);
                    break;
                case 'date':
                    $ret = html5form::checkDateData($data, $restrictions);
                    break;
                case 'email':
                    $ret = html5form::checkEmailData($data, $restrictions);
                    break;
                case 'float':
                    $ret = html5form::checkFloatData($data, $restrictions);
                    break;
                case 'iban':
                    $ret = html5form::checkIbanData($data, $restrictions);
                    break;
                case 'integer':
                    $ret = html5form::checkIntegerData($data, $restrictions);
                    break;
                case 'ip':
                    $ret = html5form::checkIpData($data, $restrictions);
                    break;
                case 'name':
                    $ret = html5form::checkNameData($data, $restrictions, $countryCode);
                    break;
                case 'password':
                    $ret = html5form::checkTextData($data, $restrictions);
                    break;
                case 'radio':
                    $ret = (strlen($data) > 0);
                    break;
                case 'tel':
                    $ret = html5form::checkTelData($data, $restrictions);
                    break;
                case 'text':
                case 'textarea':
                    $ret = html5form::checkTextData($data, $restrictions);
                    break;
                case 'time':
                    $ret = html5form::checkTimeData($data, $restrictions);
                    break;
                case 'url':
                    $ret = html5form::checkUrlData($data, $restrictions);
                    break;
                case 'zipcode':
                    $ret = html5form::checkZipCodeData($data, $restrictions, $countryCode);
                    break;
            }
        }
        return $ret;
    }

    /**
     * static function checkBicData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkBicData($data, $restrictions = array()) {
        $ret = FormValidation::isValidBic($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkCreditCardData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkCreditCardData($data, $restrictions = array()) {
        $ret = FormValidation::isValidCreditCard($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkCurrencyData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkCurrencyData($data, $restrictions = array()) {
        $ret = FormValidation::isValidCurrency($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkDateData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkDateData($data, $restrictions = array()) {
        $ret = FormValidation::isValidDate($data, 'Y-m-d');
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkEmailData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkEmailData($data, $restrictions = array()) {
        $ret = FormValidation::isValidEmailAddress($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkFloatData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkFloatData($data, $restrictions = array()) {
        $ret = FormValidation::isValidFloat($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkIbanData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkIbanData($data, $restrictions = array()) {
        $ret = FormValidation::isValidIBAN($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkIntegerData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkIntegerData($data, $restrictions = array()) {
        $ret = FormValidation::isValidInteger($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkIpData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkIpData($data, $restrictions = array()) {
        $ret = FormValidation::isValidIpAddress($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkNameData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @param string $countryCode Country code to check the zip code for
     * @return bool Is the data entered okay
     */
    protected static function checkNameData($data, $restrictions = array(), $countryCode = '') {
        $ret = FormValidation::isValidName($data, $countryCode);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkTelData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkTelData($data, $restrictions = array()) {
        $ret = FormValidation::isValidPhoneNumber($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkTextData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkTextData($data, $restrictions = array()) {
        $ret = true;
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkTimeData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkTimeData($data, $restrictions = array()) {
        $ret = FormValidation::isValidTime($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkUrlData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @return bool Is the data entered okay
     */
    protected static function checkUrlData($data, $restrictions = array()) {
        $ret = FormValidation::isValidUrl($data);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function checkZipCodeData()
     * 
     * Checks the data entered in a field for being valid
     * 
     * @param $data The data entered in the field
     * @param array $restrictions Restrictions (array(array('operator' => <Operator>, 'compare' => <Compare>), array('operator'...
     * @param string $countryCode Country code to check the zip code for
     * @return bool Is the data entered okay
     */
    protected static function checkZipCodeData($data, $restrictions = array(), $countryCode = '') {
        $ret = FormValidation::isValidZipCode($data, $countryCode);
        if ($ret) {
            foreach ($restrictions as $restriction) {
                $ret = html5form::executeOperation($restriction['operator'], $data, $restriction['compare'], $restriction['function']);
            }
        }
        return $ret;
    }

    /**
     * static function executeOperation()
     * 
     * Executes the operation and returns the result
     * 
     * @param string $operator The operator
     * @param $value1 The first value
     * @param $value2 The second value
     * @param string $function The function to execute on $value1 first
     * @return The result of the executed operation
     */
    protected static function executeOperation(string $operator, $value1, $value2, string $function = '') {
        if (strlen($function)) {
            if ($function == 'str_word_count') {
                $function = 'wordCount';
            }
            if (is_callable($function)) {
                $value1 = ($function)($value1);
            }
        }
        switch ($operator) {
            case '==': return ($value1 == $value2);
            case '===': return ($value1 === $value2);
            case '!=': return ($value1 != $value2);
            case '!==': return ($value1 !== $value2);
            case '>': return ($value1 > $value2);
            case '>=': return ($value1 >= $value2);
            case '<': return ($value1 < $value2);
            case '<=': return ($value1 <= $value2);
        }
    }
}

function wordCount(string $string) {
    $t = array(' ', "\t", '=', '+', '-', '*', '/', '\\', ',', '.', ';', ':', '[', ']', '{', '}', '(', ')', '<', '>', '&', '%', '$', '@', '#', '^', '!', '?', '~'); // separators
    $string = trim(str_replace($t, ' ', $string));
    $string = trim(preg_replace("/\s+/", ' ', $string));
    $num = 0;
    if (mb_strlen($string, 'UTF-8')) {
        $word_array = explode(' ', $string);
        $num = count($word_array);
    }
    return $num;
}
?>